#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright Toolkit Auhtors

_extra_supports = {
    'cassandra': True
}

import os
import re

import pandas as pd
from sqlalchemy import sql
from sqlalchemy.types import VARCHAR

try:
    from cassandra.auth import PlainTextAuthProvider
    from cassandra.cluster import Cluster
    from cassandra.policies import WhiteListRoundRobinPolicy
    import pandra as cql
except ImportError:
    _extra_supports['cassandra'] = False

from pydtk.utils.utils import load_config

from . import BaseDBHandler as _BaseDBHandler
from . import register_handler


def _replace_into(table, conn, keys, data_iter):
    from sqlalchemy.ext.compiler import compiles
    from sqlalchemy.sql.expression import Insert

    @compiles(Insert)
    def replace_string(insert, compiler, **kw):
        s = compiler.visit_insert(insert, **kw)
        s = s.replace("INSERT INTO", "INSERT OR REPLACE INTO")
        return s

    data = [dict(zip(keys, row)) for row in data_iter]

    conn.execute(table.table.insert(replace_string=""), data)


@register_handler(db_classes=['time_series'], db_engines=['sqlite', 'mysql'])
class TimeSeriesDBHandler(_BaseDBHandler):
    """Time series database handler."""

    __version__ = 'v3'
    db_defaults = load_config(__version__).sql.time_series
    _df_class = 'time_series_df'
    _df_name = 'time_series_df'
    _columns = None

    def _initialize_engine(self,
                           db_engine=None,
                           db_host=None,
                           db_name=None,
                           db_username=None,
                           db_password=None):
        """Initialize DB engine."""
        # Load settings from environment variables
        engine = db_engine if db_engine is not None \
            else os.environ.get('PYDTK_TIME_SERIES_DB_ENGINE', None)
        username = db_username if db_username is not None \
            else os.environ.get('PYDTK_TIME_SERIES_DB_USERNAME', None)
        password = db_password if db_password is not None \
            else os.environ.get('PYDTK_TIME_SERIES_DB_PASSWORD', None)
        host = db_host if db_host is not None \
            else os.environ.get('PYDTK_TIME_SERIES_DB_HOST', None)
        database = db_name if db_name is not None \
            else os.environ.get('PYDTK_TIME_SERIES_DB_DATABASE', None)

        super()._initialize_engine(engine, host, database, username, password)

    def save(self, df=None, remove_duplicates=False, **kwargs):
        """Save data to SQL.

        Args:
            df (pandas.DataFrame): DataFrame to save (if None, self.df will be saved)
            remove_duplicates (bool): if True, duplicated rows will be removed
            **kwargs: kwargs for function `pandas.dataframe.to_sql`

        """
        dataframe = df.copy() if df is not None else self.df.copy()
        dataframe = dataframe[~dataframe.index.duplicated(keep='last')]

        if 'index' not in kwargs.keys():
            kwargs.update({'index': True})
        self._prepare_columns()

        # TODO: alter column 'uuid_in_df' as primary key or wait for the PR below
        # TODO: Use `if_exists=upsert_overwrite` after the following PR is merged
        # https://github.com/pandas-dev/pandas/pull/29636
        dataframe.to_sql(self.df_name,
                         self._engine,
                         if_exists='append',
                         dtype={'uuid_in_df': VARCHAR(32)},
                         method=_replace_into,
                         **kwargs)


@register_handler(db_classes=['time_series'], db_engines=['cassandra'])
class TimeSeriesCassandraDBHandler(TimeSeriesDBHandler):
    """Handler for Time series database using Apache Cassandra as backend."""

    def __init__(self, *args, **kwargs):
        """Initialize Handler.

        Args:
            *args: args for super().__init__()
            **kwargs: kwargs for super().__init__()

        """
        super().__init__(*args, **kwargs)

    def _initialize_engine(self,
                           db_engine=None,
                           db_host=None,
                           db_name=None,
                           db_username=None,
                           db_password=None):
        """Initialize DB engine."""
        if not _extra_supports['cassandra']:
            raise ImportError('Cassandra-related modules could not be loaded.',
                              'Make sure you installed extra modules')

        # Load settings from environment variables
        engine = db_engine if db_engine is not None else os.environ.get('PYDTK_TIME_SERIES_DB_ENGINE', None)
        username = db_username if db_username is not None else os.environ.get('PYDTK_TIME_SERIES_DB_USERNAME', None)
        password = db_password if db_password is not None else os.environ.get('PYDTK_TIME_SERIES_DB_PASSWORD', None)
        host = db_host if db_host is not None else os.environ.get('PYDTK_TIME_SERIES_DB_HOST', None)
        database = db_name if db_name is not None else os.environ.get('PYDTK_TIME_SERIES_DB_DATABASE', None)

        # Load default settings from config file
        engine = engine if engine is not None else self.db_defaults.engine
        username = username if username is not None else self.db_defaults.username
        password = password if password is not None else self.db_defaults.password
        host = host if host is not None else self.db_defaults.host
        database = database if database is not None else self.db_defaults.database

        assert engine == 'cassandra'

        # Substitute
        self._config.current_db = {
            'engine': engine,
            'host': host,
            'username': username,
            'password': password,
            'database': database,
        }

        # Connect
        hostname = host
        hostport = 9042
        if ':' in host:
            hostname, hostport = host.split(':')

        auth_provider = PlainTextAuthProvider(username=username, password=password)
        cluster = Cluster(contact_points=[hostname],
                          port=int(hostport),
                          connect_timeout=int(os.environ.get('PYDTK_CASSANDRA_CONNECT_TIMEOUT', 60)),
                          auth_provider=auth_provider,
                          load_balancing_policy=WhiteListRoundRobinPolicy([hostname]))

        # Define pandas factory function
        def pandas_factory(colnames, rows):
            return pd.DataFrame(rows, columns=colnames)

        session = cluster.connect(database)
        session.row_factory = pandas_factory
        session.default_fetch_size = None

        self._engine = cluster
        self._session = session

    def read(self, df_name=None, query=None, where=None, order_by=None, **kwargs):
        """Read data from SQL.

        Args:
            df_name (str): Dataframe name to read
            query (str SQL query or SQLAlchemy Selectable): query to select items
            where (str): query string for filtering items
            order_by (srt): column name to sort by
            **kwargs: kwargs for function `pandas.read_sql_query`
                      or `influxdb.DataFrameClient.query`

        Returns:
            (pandas.df): table

        """
        if df_name is not None:
            self.df_name = df_name

        # Create a query
        q = query
        if q is None:
            # Check if the table exits on DB
            temp = self._session.execute(
                "select table_name from system_schema.tables "
                "where keyspace_name = '{0}' and table_name = '{1}'"
                .format(self._config.current_db['database'], self._df_name)
            )
            if len(temp._current_rows) == 0:
                self.df = self._initialize_df()
                self.logger.warning('Table "{0}" not found on DB'.format(str(df_name)))
                return

            # Create a sub-query for extracting unique records
            sub_q = sql.select('*', from_obj=sql.table(self._df_name))

            # Create a query
            q = sub_q
            if where is not None:
                q = q.where(sql.text(where))
            if order_by is not None:
                q = q.order_by(sql.text(order_by))

        # Read table from DB
        try:
            res = self._session.execute(str(q), timeout=None)
            df = res._current_rows
        except Exception as e:
            df = self._initialize_df()
            self.logger.warning('Could not execute SQL statement: "{0}" (reason: {1})'.
                                format(str(q), str(e)))

        self.df = df

    def save(self, df=None, **kwargs):
        """Save data to SQL.

        Args:
            df (pandas.DataFrame): dataframe to save. if None, self.df will be saved

        """
        dataframe: pd.DataFrame = df if df is not None else self.df

        # Convert to CassandraDataFrame
        cql_df = cql.CassandraDataFrame(dataframe)

        # Get column types
        primary_key = ['timestamp']
        if 'record_id' in [c['name'] for c in self.columns]:
            primary_key = ['record_id'] + primary_key
        column_types = cql_df._infer_data_type_from_dtype(primary_key)
        for value in column_types.values():
            value.column_type = 'double' if value.column_type == 'float' else value.column_type
            value.column_type = 'double' if value.column_type == 'int' else value.column_type
            value.name = '"{}"'.format(value.name)

        # Prepare columns
        self._prepare_columns(column_types)

        # Write to DB
        cql_df.to_cassandra(cassandra_session=self._session,
                            table_name='"{}"'.format(self._df_name),
                            data_types=column_types,
                            create_table=True)

    def _prepare_columns(self, column_types):
        """Add missing column to the table on DB.

        Args:
            column_types (list of dict): list of column name and type

        """
        # Check if the table exits on DB
        temp = self._session.execute(
            "select table_name from system_schema.tables "
            "where keyspace_name = '{0}' and table_name = '{1}'"
            .format(self._config.current_db['database'], self._df_name)
        )
        if len(temp._current_rows) == 0:
            return

        # Get columns in the table on DB
        db_table_info = self._session.execute(
            "select column_name from system_schema.columns "
            "where keyspace_name = '{0}' "
            " and table_name = '{1}' allow filtering"
            .format(self._config.current_db['database'], self._df_name)
        )._current_rows.to_dict('list')

        # Calculate difference
        difference = set(column_types.keys()).difference(db_table_info['column_name'])

        # Add columns
        for new_column in difference:
            self._session.execute(
                'ALTER TABLE {keyspace_name}.{table_name} ADD "{column_name}" {dtype};'
                .format(
                    keyspace_name=self._config.current_db['database'],
                    table_name=self._df_name,
                    column_name=new_column,
                    dtype=column_types[new_column].column_type
                )
            )

    @property
    def df_name(self):
        """Return df_name."""
        value = self._df_name
        return value

    @df_name.setter
    def df_name(self, value):
        """Setter for self.df_name."""
        if not re.fullmatch('[a-zA-Z_0-9]+', value):
            raise ValueError(
                'Invalid df name: {}  '
                'df name can only contain alphabet, number and underscore'
            )
        self._df_name = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright Toolkit Authors

import hashlib
import os

from .time_series import TimeSeriesDBHandler
from .time_series import TimeSeriesCassandraDBHandler
from . import register_handler


@register_handler(db_classes=['statistics'], db_engines=['sqlite', 'mysql', 'mariadb'])
class StatisticsDBHandler(TimeSeriesDBHandler):
    """DB Handler for statistics data."""

    _database_id: str = ''
    _span = None
    _df_class = 'statistics_df'
    _df_name = 'statistics_df'

    def __init__(self, database_id: str, span: float, **kwargs):
        """Initialize StatisticsDBHandler.

        Args:
            database_id (str): ID of the database
            span (float): interval of statistic values

        """
        self._database_id = database_id
        self._span = span
        super(StatisticsDBHandler, self).__init__(**kwargs)

    def _initialize_engine(self,
                           db_engine=None,
                           db_host=None,
                           db_name=None,
                           db_username=None,
                           db_password=None):
        """Initialize DB engine."""
        # Load settings from environment variables
        engine = db_engine if db_engine is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_ENGINE', None)
        username = db_username if db_username is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_USERNAME', None)
        password = db_password if db_password is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_PASSWORD', None)
        host = db_host if db_host is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_HOST', None)
        database = db_name if db_name is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_DATABASE', None)

        super()._initialize_engine(engine, host, database, username, password)

    @property
    def _df_name(self):
        """Return _df_name."""
        template = self._config.statistics_df.df_name
        database_id_hashed = hashlib.blake2s(
            self._database_id.encode('utf-8'),
            digest_size=self._config.hash.digest_size
        ).hexdigest()
        return template.format(**{
            'database_id': database_id_hashed,
            'span': self._span
        })

    @_df_name.setter
    def _df_name(self, value):
        """Setter for self._df_name."""
        raise RuntimeError(
            'Setting df_name is not supported in StatisticsDBHandler'
        )


@register_handler(db_classes=['statistics'], db_engines=['cassandra'])
class StatisticsCassandraDBHandler(TimeSeriesCassandraDBHandler):
    """DB Handler for statistics data using Apache Cassandra."""

    _database_id: str = ''
    _span: float = None
    _df_class = 'statistics_df'
    _df_name = 'statistics_df'

    def __init__(self, database_id: str, span: float, **kwargs):
        """Initialize StatisticsDBHandler.

        Args:
            database_id (str): ID of the database
            span (float): interval of statistic values

        """
        self._database_id = database_id
        self._span = span
        super(StatisticsCassandraDBHandler, self).__init__(**kwargs)

    def _initialize_engine(self,
                           db_engine=None,
                           db_host=None,
                           db_name=None,
                           db_username=None,
                           db_password=None):
        """Initialize DB engine."""
        # Load settings from environment variables
        engine = db_engine if db_engine is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_ENGINE', None)
        username = db_username if db_username is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_USERNAME', None)
        password = db_password if db_password is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_PASSWORD', None)
        host = db_host if db_host is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_HOST', None)
        database = db_name if db_name is not None \
            else os.environ.get('PYDTK_STATISTICS_DB_DATABASE', None)

        super()._initialize_engine(engine, host, database, username, password)

    @property
    def _df_name(self):
        """Return _df_name."""
        template = self._config[self._df_class]['df_name']
        database_id_hashed = hashlib.blake2s(
            self._database_id.encode('utf-8'),
            digest_size=self._config.hash.digest_size
        ).hexdigest()
        return template.format(**{
            'database_id': database_id_hashed,
            'span': self._span
        })

    @_df_name.setter
    def _df_name(self, value):
        """Setter for self._df_name."""
        raise RuntimeError(
            'Setting df_name is not supported in StatisticsDBHandler'
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""DB Engines for V4DBHandler."""

from datetime import datetime
import logging
import os
from typing import Optional

from tinydb import TinyDB
from tinydb import __version__ as tinydb_version
from tinydb.database import Document as _Document
from tinydb.database import StorageProxy as _StorageProxy
from tinymongo import TinyMongoClient
from ..deps import pql as PQL

logger = logging.getLogger(__name__)


DEFAULT_DB_NAME = 'default'
DEFAULT_COLLECTION_NAME = 'default'


class Document(_Document):
    """Custom Document."""

    def __getitem__(self, item):
        """Get item.

        Args:
            item (str): item-name

        Returns:
            (Document)

        """
        if '.' not in item:
            return super().__getitem__(item)
        else:
            current_item = self
            for sub_item in item.split('.'):
                current_item = current_item.__getitem__(sub_item)
            return current_item


class StorageProxy(_StorageProxy):
    """Custom StorageProxy."""

    def _new_document(self, key, val):
        doc_id = int(key)
        return Document(val, doc_id)


def connect(
    db_host: str,
    db_name: Optional[str] = None,
    collection_name: Optional[str] = None,
    **kwargs
):
    """Connect to DB.

    Args:
        db_host (str): database host
        db_name (str): database name
        collection_name (str): collection name

    Returns:
        (TinyMongoCollection): connection

    """
    if db_name is None:
        db_name = DEFAULT_DB_NAME
    if collection_name is None:
        collection_name = DEFAULT_COLLECTION_NAME

    if not os.path.isdir(db_host):
        os.makedirs(db_host, exist_ok=True)

    # Customize storage-proxy
    TinyDB.storage_proxy_class = StorageProxy

    connection = TinyMongoClient(db_host)
    db = getattr(connection, db_name)
    collection = getattr(db, collection_name)

    # Add attributes
    collection._db_host = db_host
    collection._db_name = db_name
    collection._collection_name = collection_name

    return collection


def read(db,
         query: Optional[dict] = None,
         pql: any = None,
         order_by: Optional[list] = None,
         **kwargs):
    """Read data from DB.

    Args:
        db (TinyMongoCollection): DB connection
        query (dict or Query): Query to select items
        pql (PQL) Python-Query-Language to select items
        order_by (list): column name to sort by with format [ ( column1, 1 or -1 ), ... ]
        **kwargs: kwargs for function `pandas.read_sql_query`
                  or `influxdb.DataFrameClient.query`

    Returns:
        (list, int): list of data and total number of records

    """
    # Re-initialize DB to reload data (This operation is needed as tinymongo caches data in memory)
    db = connect(
        db_host=db._db_host,
        db_name=db._db_name,
        collection_name=db._collection_name,
    )

    if pql is not None and query is not None:
        raise ValueError('Either query or pql can be specified')

    if pql:
        query = PQL.find(pql)

    # Fix query
    if query:
        query = _fix_query_exists(query)

    if query:
        if order_by is None:
            data = db.find(query)
        else:
            data = db.find(query).sort(order_by)
    else:
        if order_by is None:
            data = db.find()
        else:
            data = db.find().sort(order_by)

    data = list(data)

    return data, len(data)


def upsert(db, data, **kwargs):
    """Write data to DB.

    Args:
        db (TinyMongoCollection): DB connection
        data (list): data to save

    """
    for record in data:
        _record = _fix_datetime(record)
        uuid = record['_uuid']
        if db.find_one({'_uuid': uuid}) is not None:
            db.update({'_uuid': uuid}, _record)
        else:
            db.insert(_record)


def remove(db, uuids, **kwargs):
    """Remove data from DB.

    Args:
        db (TinyMongoCollection): DB connection
        uuids (list): A list of unique IDs

    """
    for uuid in uuids:
        db.delete_many({'_uuid': uuid})


def drop_table(db, name, **kwargs):
    """Drop a table from DB.

    Args:
        db (TinyMongoCollection): DB connection
        name (str): Name of the target table

    """
    if tinydb_version.startswith('4'):
        db.parent.tinydb.drop_table(name)
    else:
        db.parent.tinydb.purge_table(name)


def exist_table(db, name, **kwargs):
    """Check if the specified table (collection) exist.

    Args:
        db (TinyMongoCollection): DB connection
        name (str): Name of the target table

    """
    return name in list(db.parent.tinydb.tables())


def _fix_query_exists(query):
    if isinstance(query, list):
        fixed_query = []
        for item in query:
            fixed_query.append(_fix_query_exists(item))
        return fixed_query

    elif isinstance(query, dict):
        fixed_query = {}
        for key, value in query.items():
            if isinstance(value, dict) or isinstance(value, list):
                fixed_query[key] = _fix_query_exists(value)
            elif key == '$exists' and value:
                fixed_query['$ne'] = None
            else:
                fixed_query[key] = value

        return fixed_query

    else:
        raise ValueError


def _fix_datetime(data: dict):
    """Fix datetime object.

    Args:
        data (dict): Input data

    Returns:
        (dict): Fixed data

    """
    _data = {}
    for key, value in data.items():
        if isinstance(value, datetime):
            _data[key] = value.timestamp()
        else:
            _data[key] = value

    return _data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""Base preprocess modules."""

from abc import ABCMeta


class BasePreprocess(metaclass=ABCMeta):
    """Base Preprocess."""

    def __init__(self):
        pass

    def processing(self, timestamps, data):
        """Process signals."""
        raise NotImplementedError()



from datetime import datetime
from unittest import TestCase
import bson
import pql

class BasePqlTestCase(TestCase):

    def compare(self, string, expected):
        print("{} | {}".format(string,expected))
        self.assertEqual(pql.find(string), expected)

class PqlSchemaLessTestCase(BasePqlTestCase):

    def test_hyphenated(self):
        self.compare('"foo-bar" == "spam"', {'foo-bar': 'spam'})

    def test_equal_int(self):
        self.compare('a == 1', {'a': 1})

    def test_not_equal_string(self):
        self.compare('a != "foo"', {'a': {'$ne': 'foo'}})

    def test_nested(self):
        self.compare('a.b == 1', {'a.b': 1})

    def test_and(self):
        self.compare('a == 1 and b == 2', {'$and': [{'a': 1}, {'b': 2}]})

    def test_or(self):
        self.compare('a == 1 or b == 2', {'$or': [{'a': 1}, {'b': 2}]})

    def test_not(self):
        self.compare('not a > 1', {'a': {'$not': {'$gt': 1}}})

    def test_algebra(self):
        for string, expected in [('a > 1', {'a': {'$gt': 1}}),
                                 ('a >= 1', {'a': {'$gte': 1}}),
                                 ('a < 1', {'a': {'$lt': 1}}),
                                 ('a <= 1', {'a': {'$lte': 1}})]:
            self.compare(string, expected)

    def test_bool(self):
        self.compare('a == True', {'a': True})
        self.compare('a == False', {'a': False})

    def test_none(self):
        self.compare('a == None', {'a': None})
        self.compare('a == null', {'a': None})

    def test_list(self):
        self.compare('a == [1, 2, 3]', {'a': [1, 2, 3]})

    def test_dict(self):
        self.compare('a == {"foo": 1}', {'a': {'foo': 1}})

    def test_in(self):
        self.compare('a in [1, 2, 3]', {'a': {'$in': [1, 2, 3]}})

        with self.assertRaises(pql.ParseError) as context:
            pql.find('a in (1)')

        self.assertIn('Invalid value type', str(context.exception))

    def test_not_in(self):
        self.compare('a not in [1, 2, 3]', {'a': {'$nin': [1, 2, 3]}})

        with self.assertRaises(pql.ParseError) as context:
            pql.find('a not in (1)')

        self.assertIn('Invalid value type', str(context.exception))

    def test_missing_func(self):
        with self.assertRaises(pql.ParseError) as context:
            pql.find('a == foo()')
        self.assertIn('Unsupported function', str(context.exception))

    def test_invalid_name(self):
        with self.assertRaises(pql.ParseError) as context:
            pql.find('a == foo')
        self.assertIn('Invalid name', str(context.exception))

    def test_exists(self):
        self.compare('a == exists(True)', {'a': {'$exists': True}})

    def test_type(self):
        self.compare('a == type(3)', {'a': {'$type': 3}})

    def test_regex(self):
        self.compare('a == regex("foo")', {'a': {'$regex': 'foo'}})
        self.compare('a == regex("foo", "i")', {'a': {'$regex': 'foo', '$options': 'i'}})

    def test_mod(self):
        self.compare('a == mod(10, 3)', {'a': {'$mod': [10, 3]}})

    def test_size(self):
        self.compare('a == size(4)', {'a': {'$size': 4}})

    def test_all(self):
        self.compare('a == all([1, 2, 3])', {'a': {'$all': [1, 2, 3]}})

    def test_match(self):
        self.compare('a == match({"foo": "bar"})', {'a': {'$elemMatch': {'foo': 'bar'}}})

    def test_date(self):
        self.compare('a == date(10)', {'a': datetime.fromtimestamp(10)})
        self.compare('a == date("2012-3-4")', {'a': datetime(2012, 3, 4)})
        self.compare('a == date("2012-3-4 12:34:56.123")',
                     {'a': datetime(2012, 3, 4, 12, 34, 56, 123000)})

    def test_epoch(self):
        self.compare('a == epoch(10)', {'a': 10})
        self.compare('a == epoch("2012/1/1")', {'a': 1325372400.0})

    def test_epoch_utc(self):
        self.compare('a == epoch_utc(10)', {'a': 10})
        self.compare('a == epoch_utc("2012/1/1")', {'a': 1325376000.0})

    def test_id(self):
        self.compare('_id == id("abcdeabcdeabcdeabcdeabcd")',
                     {'_id': bson.ObjectId("abcdeabcdeabcdeabcdeabcd")})

    def test_near_legacy_coordinates(self):
        self.compare('location == near([1, 2], 10)',
                     {'location':
                      {'$near': [1,2],
                       '$maxDistance': 10}})

    def test_near_sphere_point(self):
        self.compare('location == nearSphere(Point(1, 2))',
                     {'location':
                      {'$nearSphere':
                       {'$geometry':
                        {'type': 'Point',
                         'coordinates': [1, 2]}}}})

    def test_near_point_with_max(self):
        self.compare('location == near(Point(1, 2), 10)',
                     {'location':
                      {'$near':
                       {'$geometry':
                        {'type': 'Point',
                         'coordinates': [1, 2]},
                        '$maxDistance': 10}}})

    def test_geo_within_polygon(self):
        polygon = [[1, 2], [3, 4], [5, 6]]
        self.compare('location == geoWithin(Polygon([{}, {}]))'.format(polygon, polygon),
                     {'location':
                      {'$geoWithin':
                       {'$geometry':
                        {'type': 'Polygon',
                         'coordinates': [polygon, polygon]}}}})

    def test_geo_intersects_line_string(self):
        self.compare('location == geoIntersects(LineString([[1, 2], [3, 4]]))',
                     {'location':
                      {'$geoIntersects':
                       {'$geometry':
                        {'type': 'LineString',
                         'coordinates': [[1, 2], [3, 4]]}}}})

    def test_center_within(self):
        for center_type in ['center', 'centerSphere']:
            self.compare('location == geoWithin({}([1, 2], 3))'.format(center_type),
                         {'location':
                          {'$geoWithin':
                           {'$' + center_type: [[1, 2], 3]}}})

    def test_polygon_and_box(self):
        for shape in ['box', 'polygon']:
            self.compare('location == geoWithin({}([[1, 2], [3, 4], [5, 6]]))'.format(shape),
                         {'location':
                          {'$geoWithin':
                           {'$' + shape: [[1, 2], [3, 4], [5, 6]]}}})

class PqlSchemaAwareTestCase(BasePqlTestCase):

    def compare(self, string, expected):
        print("{} | {}".format(string,expected))
        self.assertEqual(pql.find(string, schema={'a': pql.IntField(),
                                                  'd': pql.DateTimeField(),
                                                  'foo.bar': pql.ListField(pql.StringField())}), expected)

    def test_sanity(self):
        self.compare('a == 3', {'a': 3})

    def test_invalid_field(self):
        with self.assertRaises(pql.ParseError) as context:
            self.compare('b == 3', None)
        self.assertEqual(sorted(context.exception.options),
                         sorted(['a', 'd', 'foo.bar']))

    def test_type_error(self):
        with self.assertRaises(pql.ParseError):
            self.compare('a == "foo"', None)

    def test_invalid_function(self):
        with self.assertRaises(pql.ParseError) as context:
            self.compare('a == size(3)', None)
        self.assertIn('Unsupported function', str(context.exception))

    def test_invalid_date(self):
        with self.assertRaises(pql.ParseError) as context:
            self.compare('d == "foo"', None)
        self.assertIn('Error parsing date', str(context.exception))

    def test_date(self):
        self.compare('d > "2012-03-02"',
                     {'d': {'$gt': datetime(2012, 3, 2)}})

    def test_nested(self):
        self.compare('foo.bar == ["spam"]', {'foo.bar': ['spam']})
        #self.compare('foo.bar == "spam"', {'foo.bar': 'spam'}) # currently broken



#!/usr/bin/env python3

# Copyright Toolkit Authors (Yusuke Adachi)

import abc
import numpy as np
import rosbag

import time  # for debug


class BaseReader(object, metaclass=abc.ABCMeta):
    """Base class for data reader."""

    def __init__(self, file):
        """Set file."""
        self.file = file

    @abc.abstractmethod
    def contents(self):
        """Return ndarray data."""
        pass


class RosbagReader(BaseReader):
    """Rosbag data reader class."""

    def __init__(self, file):
        """Set file."""
        super(RosbagReader, self).__init__(file)

    def _get_msg_type(self, msg):
        try:
            _ = msg.data
            return "std_msgs/"
        except AttributeError:
            try:
                _ = msg.latitude
                return "sensor_msgs/NavSatFix"
            except AttributeError:
                try:
                    _ = msg.accel
                    return "geometry_msgs/AccelStamped"
                except AttributeError:
                    try:
                        _ = msg.range
                        return "sensor_msgs/Range"
                    except AttributeError:
                        raise Exception('Unsupported message type: %s' % type(msg))

    def _msg_to_list(self, msg, msg_type="std_msgs/"):
        if msg_type == "std_msgs/":
            return [msg.data]
        elif msg_type == "sensor_msgs/NavSatFix":
            return [msg.latitude, msg.longitude]
        elif msg_type == "geometry_msgs/AccelStamped":
            return [msg.accel.linear.x, msg.accel.linear.y, msg.accel.linear.z]
        elif msg_type == "sensor_msgs/Range":
            return [msg.range]

    def contents(self, content):
        """Return timestamps and data(ndarray)."""
        bag = rosbag.Bag(self.file)
        time_list, data_list = None, None
        for topic, msg, t in bag.read_messages():
            if topic == content:
                timestamp = t.secs + float(t.nsecs) / (10 ** 9)
                if data_list is None:
                    time_list = [[timestamp]]
                    msg_type = self._get_msg_type(msg)
                    data_list = [self._msg_to_list(msg, msg_type)]
                else:
                    time_list.append([timestamp])
                    data_list.append(self._msg_to_list(msg, msg_type))
        return np.array(time_list), np.array(data_list)


if __name__ == '__main__':
    reader = RosbagReader("/data_pool_1/small_DrivingBehaviorDatabase/records/ \
                          016_00000000030000000240/data/records.bag")

    t1 = time.time()

    test = reader.contents("/vehicle/analog/speed_pulse")

    print(test.shape)
    print(test[0])
    print(time.time() - t1)



#!/usr/bin/env python3

# Copyright Toolkit Authors
import json
import os
import sys

import pandas
from pydtk.db import DBHandler
from pydtk.models import MetaDataModel

pandas.set_option('display.max_columns', None)
pandas.set_option('display.width', None)


class EmptySTDINError(Exception):
    """Exception for the case that STDIN is empty."""

    pass


def _get_db_handler(target: str, database_id: str = 'default', base_dir: str = '/'):
    if target in ['databases', 'database']:
        handler = DBHandler(
            db_class='database_id'
        )
        group_by = None
    elif target in ['records', 'record']:
        handler = DBHandler(
            db_class='meta',
            database_id=database_id,
            base_dir_path=base_dir,
            orient='record_id'
        )
        group_by = 'record_id'
    elif target in ['files', 'file']:
        handler = DBHandler(
            db_class='meta',
            database_id=database_id,
            base_dir_path=base_dir,
            orient='path'
        )
        group_by = None
    elif target in ['contents', 'content']:
        handler = DBHandler(
            db_class='meta',
            database_id=database_id,
            base_dir_path=base_dir,
            orient='contents'
        )
        group_by = None
    else:
        raise ValueError('Unknown target')
    return handler, group_by


def _assert_target(target):
    assert target in [
        'database',
        'databases',
        'record',
        'records',
        'file',
        'files',
        'content',
        'contents'
    ], \
        "target must be one of 'database', 'record', 'file' or 'content"


def _add_data_from_stdin(handler):
    raw_data = sys.stdin.read()
    if raw_data == '':
        raise EmptySTDINError('STDIN is empty')
    data = json.loads(raw_data)
    if isinstance(data, dict):
        metadata = MetaDataModel(data=data)
        handler.add_data(metadata.data)
    elif isinstance(data, list):
        for element in data:
            metadata = MetaDataModel(data=element)
            handler.add_data(metadata.data)


class DB(object):
    """DB."""

    _handler = None

    def list(
        self,
        target: str,
        database_id: str = 'default',
        pql: str = None,
        offset: int = 0,
        limit: int = 20,
        order_by: str = None,
        base_dir: str = '/',
        **kwargs
    ):
        """List resources in DB.

        Args:
            target (str): 'databases', 'records', 'files' or 'contents
            *
            database_id (str): Database ID
            pql (str): Python-Query-Language for searching metadata
            offset (int): Cursor offset
            limit (int): Number of elements to show
            order_by (str): Sort key
            base_dir (str): Base directory

        Examples:
            List available databases using CLI:

            ```bash
            $ pydtk db list databases

            ```

            List records:
            ```bash
            $ pydtk db list records --database_id <Database ID>

            ```

            List records with PQL:
            ```bash
            $ pydtk db list records --database_id <Database ID> --pql 'record_id == regex("abc.*")'

            ```

            List records and display as a parsable string
            ```bash
            $ pydtk db list records --parsable

            ```

        """
        _assert_target(target)

        handler, group_by = _get_db_handler(target, database_id=database_id, base_dir=base_dir)

        # Prepare search query
        if order_by is not None:
            order_by = [(order_by, 1)]

        # Read
        handler.read(pql=pql, limit=limit, offset=offset, order_by=order_by, group_by=group_by)

        # Display
        _display(
            handler,
            columns=['Database ID', 'Record ID', 'Description', 'File path', 'Contents', 'Tags'],
            **kwargs
        )

        self._handler = handler

    def get(
        self,
        target: str,
        database_id: str = 'default',
        record_id: str = None,
        path: str = None,
        content: str = None,
        base_dir: str = '/',
        **kwargs,
    ):
        """Get resources.

        Args:
            target (str): 'databases', 'records', 'files' or 'contents
            *
            database_id (str): Database ID
            record_id (str): Record ID
            path (str): File path
            content (str): Content
            base_dir (str): Base directory

        """
        _assert_target(target)

        handler, group_by = _get_db_handler(target, database_id=database_id, base_dir=base_dir)

        # Prepare query
        pql = ''
        if target in ['database', 'databases']:
            if database_id is not None:
                pql += ' and ' if pql != '' else ''
                pql += 'database_id == "{}"'.format(database_id)
        else:
            if record_id is not None:
                pql += ' and ' if pql != '' else ''
                pql += 'record_id == "{}"'.format(record_id)
            if path is not None:
                relative_path = handler._solve_path(path, 'relative')
                pql += ' and ' if pql != '' else ''
                pql += ' path == "{}"'.format(relative_path)
            if content is not None:
                pql += ' and ' if pql != '' else ''
                pql += '"contents.{}" == exists(True)'.format(content)

        # Read
        handler.read(pql=pql, group_by=group_by)

        # Display
        _display(handler, **kwargs)

        self._handler = handler

    def add(
        self,
        target: str,
        content: str = None,
        database_id: str = 'default',
        base_dir: str = '/',
        overwrite: bool = False,
        skip_checking_existence: bool = False,
        **kwargs,
    ):
        """Add resources.

        Args:
            target (str): 'databases', 'records', 'files' or 'contents'
            *
            content (str): Content to add. This must be one of the followings:
                           1. Database ID (in case of adding database)
                           2. Path to a JSON file (in case of adding metadata)
                           3. Path to a directory containing JSON files (in case of adding metadata)
                           4. Empty (in case of adding metadata)
                           In the last case, PyDTK reads STDIN as JSON to add metadata.
            database_id (str): Database ID
            base_dir (str): Base directory
            overwrite (bool): Overwrite the existing data on DB
            skip_checking_existence (bool): Skip checking the existence of the input data in DB

        """
        _assert_target(target)

        num_added = 0
        num_updated = 0

        # Initialize Handler
        handler, _ = _get_db_handler(target, database_id=database_id, base_dir=base_dir)

        if target in ['database', 'databases']:
            database_id = database_id if content is None else content
            data = {
                'database_id': database_id,
                'name': database_id
            }
            if content is None:
                try:
                    _add_data_from_stdin(handler)
                except EmptySTDINError:
                    handler.add_data(data)
            else:
                handler.add_data(data)

        else:
            if content is None:
                _add_data_from_stdin(handler)
            else:
                if os.path.isfile(content):
                    f = open(content, 'r')
                    data = json.load(f)
                    f.close()
                    metadata = MetaDataModel(data=data)
                    handler.add_data(metadata.data)
                elif os.path.isdir(content):
                    from pydtk.builder.meta_db import _find_json as find_json
                    json_list = find_json(content)
                    for json_file in json_list:
                        metadata = MetaDataModel()
                        metadata.load(json_file)
                        handler.add_data(metadata.data)
                else:
                    raise IOError('No such file or directory')

        if not skip_checking_existence:
            # Check if the UUIDs already exist on DB
            uuids = [data['_uuid'] for data in handler.data]
            self.list(
                target,
                database_id=database_id,
                pql=' or '.join(['_uuid == "{}"'.format(uuid) for uuid in uuids]),
                quiet=True,
                include_summary=False,
            )

            # Ask
            if not overwrite and len(self._handler) > 0:
                print('The following data on DB will be overwritten.')
                _display(self._handler, **kwargs)
                sys.stdin = open('/dev/tty')
                judge = input("Proceed? [y/N]: ")
                if judge not in ['y', 'Y']:
                    print('Cancelled.')
                    sys.exit(1)

            num_updated += len(self._handler)

        num_added += len(handler) - num_updated

        # Save
        handler.save()

        # Display
        if num_updated > 0:
            print('Updated: {} items.'.format(num_updated))
        if num_added > 0:
            print('Added: {} items.'.format(num_added))

        self._handler = handler

    def delete(
        self,
        target: str,
        database_id: str = 'default',
        record_id: str = None,
        path: str = None,
        content: str = None,
        base_dir: str = '/',
        **kwargs
    ):
        """Delete resources.

        Args:
            target (str): 'database' or 'metadata'
            *
            database_id (str): Database ID
            record_id (str): Record ID
            path (str): File path
            content (str): Content
            base_dir (str): Base directory

        """
        _assert_target(target)
        yes = False
        quiet = False
        if 'y' in kwargs and kwargs['y']:
            yes = True
        if 'yes' in kwargs and kwargs['yes']:
            yes = True
        if 'q' in kwargs and kwargs['q']:
            quiet = True
        if 'quiet' in kwargs and kwargs['quiet']:
            quiet = True

        # Get the corresponding resources
        if not all([
            record_id is None,
            path is None,
            content is None
        ]):
            self.get(
                target=target,
                database_id=database_id,
                record_id=record_id,
                path=path,
                content=content,
                columns=['Database ID', 'Record ID', 'Description', 'File path', 'Contents',
                         'Tags'],
                **kwargs
            )
            handler = self._handler

        else:
            handler = DBHandler(
                db_class='meta',
                database_id=database_id,
                base_dir_path=base_dir,
            )
            _add_data_from_stdin(handler)

        # Get the corresponding data from DB
        uuids = [data['_uuid'] for data in handler.data]
        self.list(
            target,
            database_id=database_id,
            pql=' or '.join(['_uuid == "{}"'.format(uuid) for uuid in uuids]),
            quiet=True,
            include_summary=False,
        )

        if not yes and not quiet and len(self._handler) > 0:
            print('The following data will be deleted:')
            _display(self._handler, **kwargs)
            sys.stdin = open('/dev/tty')
            judge = input("Proceed? [y/N]: ")
            if judge not in ['y', 'Y']:
                print('Cancelled.')
                sys.exit(1)

        if len(self._handler) == 0:
            if not quiet:
                print(f'No such {target}')
            sys.exit(1)

        # Delete metadata
        num_deleted = len(self._handler)
        for data in self._handler.data:
            handler.remove_data(data)
            self._handler.remove_data(data)

        # Save
        self._handler.save()

        # Display
        if not quiet:
            print('Deleted: {} items.'.format(num_deleted))
            if len(handler) > 0:
                print('The following {} items were not found.'.format(len(handler)))
                _display(handler, include_summary=False, **kwargs)


def _display(
    handler: DBHandler,
    columns: list = None,
    include_summary=True,
    **kwargs
):
    """Display data.

    Args:
        handler (DBHandler): Database handler
        *
        columns (list): List of columns to display
        include_summary (bool): Display summary

    """
    parsable = False
    if 'quiet' in kwargs and kwargs['quiet']:
        return
    if 'p' in kwargs and kwargs['p']:
        parsable = True
    if 'parsable' in kwargs and kwargs['parsable']:
        parsable = True

    if not parsable:
        available_columns = [column for column in handler.df.columns if not column.startswith('_')]
        if columns is not None:
            df = handler.df[[c for c in columns if c in available_columns]]
        else:
            df = handler.df[available_columns]
        if include_summary:
            print(f'Found: {handler.count_total} items.')
        if handler.count_total > len(handler):
            print(f'Only {len(handler)} items are displayed. '
                  'Please use `--limit` or `--offset` to see more.')
        print(df)

    else:
        print(json.dumps(handler.data, indent=4, default=_default_json_handler))


def _default_json_handler(o):
    return o.__str__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""V4DB."""

# Handlers
from .handlers import BaseDBHandler as DBHandler
from .handlers.meta import MetaDBHandler
from .handlers.meta import DatabaseIDDBHandler



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""V2DB."""

# Handlers
from .handlers import BaseDBHandler
from .handlers import TimeSeriesDBHandler
from .handlers.meta import MetaDBHandler

# Search engines
from .search_engines import BaseDBSearchEngine
from .search_engines.time_series import TimeSeriesDBSearchEngine



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pickle

from pydtk.db import V3MetaDBHandler as MetaDBHandler


class LoadPKL:
    """Class for database search."""

    def __init__(self, pkl):
        self.pickle = pkl
        with open(pkl, "rb") as f:
            df_dict = pickle.load(f)
            self.content_df = df_dict["content_df"]
            self.file_df = df_dict["file_df"]
            self.record_id_df = df_dict["record_id_df"]

    def get_record_id_info(self):
        """Show record id list.

        Returns:
            (list): List of record id.

        """
        record_id_list = []
        for index, row in self.record_id_df.iterrows():
            id_dict = {
                "record_id": row["record_id"],
                "duration": row["duration"],
                "start_timestamp": row["start_timestamp"],
                "end_timestamp": row["end_timestamp"],
                "tags": row["tags"]
            }
            record_id_list += [id_dict]
        return record_id_list


class LoadDB(object):
    """Class for database search."""

    def __init__(self, path_to_db):
        super(LoadDB, self).__init__()
        self.path_to_db = path_to_db
        self.db_handler = MetaDBHandler(path_to_db)

    def get_record_id_info(self):
        """Show record info.

        Returns:
            (list): List of record information.

        """
        return self.db_handler.record_id_df.to_dict('records')



"""
The parser:
1. gets and expression
2. parses it
3. handles all boolean logic
4. delegates operator and rvalue parsing to the OperatorMap

SchemaFreeOperatorMap

  supports all mongo operators for all fields.

SchemaAwareOperatorMap

  1. verifies fields exist.
  2. verifies operators are applied to fields of correct type.

currently unsupported:
1. $where - kind of intentionally against injections
2. geospatial
"""
import ast
import bson
import datetime
import dateutil.parser
import re
from calendar import timegm


def parse_date(node):
    if type(node.value) in (int, float): # it's a number!
        return datetime.datetime.fromtimestamp(node.n)
    try:
        return dateutil.parser.parse(node.value)
    except Exception as e:
        raise ParseError('Error parsing date: ' + str(e), col_offset=node.col_offset)

class AstHandler(object):

    def get_options(self):
        return [f.replace('handle_', '') for f in dir(self) if f.startswith('handle_')]

    def resolve(self, thing):
        thing_name = thing.__class__.__name__
        try:
            handler = getattr(self, 'handle_' + thing_name)
        except AttributeError:
            raise ParseError('Unsupported syntax ({0}).'.format(thing_name,
                                                              self.get_options()),
                             col_offset=thing.col_offset if hasattr(thing, 'col_offset') else None,
                             options=self.get_options())
        return handler

    def handle(self, thing):
        return self.resolve(thing)(thing)

    def parse(self, string):
        ex = ast.parse(string, mode='eval')
        return self.handle(ex.body)

class ParseError(Exception):
    def __init__(self, message, col_offset, options=[]):
        super(ParseError, self).__init__(message)
        self.message = message
        self.col_offset = col_offset
        self.options = options
    def __str__(self):
        if self.options:
            return '{0} options: {1}'.format(self.message, self.options)
        return self.message

class Parser(AstHandler):
    def __init__(self, operator_map):
        self._operator_map = operator_map

    def get_options(self):
        return self._operator_map.get_options()

    def handle_BoolOp(self, op):
        return {self.handle(op.op): list(map(self.handle, op.values))}

    def handle_And(self, op):
        '''and'''
        return '$and'

    def handle_Or(self, op):
        '''or'''
        return '$or'

    def handle_UnaryOp(self, op):
        operator = self.handle(op.operand)
        field, value = list(operator.items())[0]
        return {field: {self.handle(op.op): value}}

    def handle_Not(self, not_node):
        '''not'''
        return '$not'

    def handle_Compare(self, compare):
        if len(compare.comparators) != 1:
            raise ParseError('Invalid number of comparators: {0}'.format(len(compare.comparators)),
                             col_offset=compare.comparators[1].col_offset)
        return self._operator_map.handle(left=compare.left,
                                         operator=compare.ops[0],
                                         right=compare.comparators[0])

class SchemaFreeParser(Parser):
    def __init__(self):
        super(SchemaFreeParser, self).__init__(SchemaFreeOperatorMap())

class SchemaAwareParser(Parser):
    def __init__(self, *a, **k):
        super(SchemaAwareParser, self).__init__(SchemaAwareOperatorMap(*a, **k))

class FieldName(AstHandler):
    def handle_Constant(self, node):
        return node.value
    def handle_Str(self, node):
        return node.s
    def handle_Name(self, name):
        return name.id
    def handle_Attribute(self, attr):
        return '{0}.{1}'.format(self.handle(attr.value), attr.attr)

class OperatorMap(object):
    def resolve_field(self, node):
        return FieldName().handle(node)
    def handle(self, operator, left, right):
        field = self.resolve_field(left)
        return {field: self.resolve_type(field).handle_operator_and_right(operator, right)}

class SchemaFreeOperatorMap(OperatorMap):
    def get_options(self):
        return None
    def resolve_type(self, field):
        return GenericField()

class SchemaAwareOperatorMap(OperatorMap):
    def __init__(self, field_to_type):
        self._field_to_type = field_to_type
    def resolve_field(self, node):
        field = super(SchemaAwareOperatorMap, self).resolve_field(node)
        try:
            self._field_to_type[field]
        except KeyError:
            raise ParseError('Field not found: {0}.'.format(field),
                             col_offset=node.col_offset,
                             options=self._field_to_type.keys())
        return field

    def resolve_type(self, field):
        return self._field_to_type[field]

#---Function-Handlers---#

class Func(AstHandler):

    @staticmethod
    def get_arg(node, index):
        if index > len(node.args) - 1:
            raise ParseError('Missing argument in {0}.'.format(node.func.id),
                             col_offset=node.col_offset)
        return node.args[index]

    @staticmethod
    def parse_arg(node, index, field):
        return field.handle(Func.get_arg(node, index))

    def handle(self, node):
        try:
            handler = getattr(self, 'handle_' + node.func.id)
        except AttributeError:
            raise ParseError('Unsupported function ({0}).'.format(node.func.id),
                             col_offset=node.col_offset,
                             options=self.get_options())
        return handler(node)

    def handle_exists(self, node):
        return {'$exists': self.parse_arg(node, 0, BoolField())}

    def handle_type(self, node):
        return {'$type': self.parse_arg(node, 0, IntField())}

class StringFunc(Func):
    def handle_regex(self, node):
        result = {'$regex': self.parse_arg(node, 0, StringField())}
        try:
            result['$options'] = self.parse_arg(node, 1, StringField())
        except ParseError:
            pass
        return result

class IntFunc(Func):
    def handle_mod(self, node):
        return {'$mod': [self.parse_arg(node, 0, IntField()),
                         self.parse_arg(node, 1, IntField())]}

class ListFunc(Func):
    def handle_size(self, node):
        return {'$size': self.parse_arg(node, 0, IntField())}

    def handle_all(self, node):
        return {'$all': self.parse_arg(node, 0, ListField())}

    def handle_match(self, node):
        return {'$elemMatch': self.parse_arg(node, 0, DictField())}

class DateTimeFunc(Func):
    def handle_date(self, node):
        return parse_date(self.get_arg(node, 0))

class IdFunc(Func):
    def handle_id(self, node):
        return self.parse_arg(node, 0, IdField())

class EpochFunc(Func):
    def handle_epoch(self, node):
        return self.parse_arg(node, 0, EpochField())

class EpochUTCFunc(Func):
    def handle_epoch_utc(self, node):
        return self.parse_arg(node, 0, EpochUTCField())

class GeoShapeFuncParser(Func):

    def handle_Point(self, node):
        return {'$geometry':
                {'type': 'Point',
                 'coordinates': [self.parse_arg(node, 0, IntField()),
                                                  self.parse_arg(node, 1, IntField())]}}

    def handle_LineString(self, node):
        return {'$geometry':
                {'type': 'LineString',
                 'coordinates': self.parse_arg(node, 0, ListField(ListField(IntField())))}}

    def handle_Polygon(self, node):
        return {'$geometry':
                {'type': 'Polygon',
                'coordinates': self.parse_arg(node, 0, ListField(ListField(ListField(IntField()))))}}

    def handle_box(self, node):
        return {'$box': self.parse_arg(node, 0, ListField(ListField(IntField())))}

    def handle_polygon(self, node):
        return {'$polygon': self.parse_arg(node, 0, ListField(ListField(IntField())))}

    def _any_center(self, node, center_name):
        return {center_name: [self.parse_arg(node, 0, ListField(IntField())),
                              self.parse_arg(node, 1, IntField())]}

    def handle_center(self, node):
        return self._any_center(node, '$center')

    def handle_centerSphere(self, node):
        return self._any_center(node, '$centerSphere')

class GeoShapeParser(AstHandler):
    def handle_Call(self, node):
        return GeoShapeFuncParser().handle(node)
    def handle_List(self, node):
        '''
        This is a legacy coordinate pair. consider supporting box, polygon, center, centerSphere
        '''
        return ListField(IntField()).handle(node)

class GeoFunc(Func):
    def _any_near(self, node, near_name):
        shape = GeoShapeParser().handle(self.get_arg(node, 0))
        result = bson.SON({near_name: shape}) # use SON because mongo expects the command before the arguments
        if len(node.args) > 1:
            distance = self.parse_arg(node, 1, IntField()) # meters
            if isinstance(shape, list): # legacy coordinate pair
                result['$maxDistance'] = distance
            else:
                shape['$maxDistance'] = distance
        return result

    def handle_near(self, node):
        return self._any_near(node, '$near')

    def handle_nearSphere(self, node):
        return self._any_near(node, '$nearSphere')

    def handle_geoIntersects(self, node):
        return {'$geoIntersects': GeoShapeParser().handle(self.get_arg(node, 0))}

    def handle_geoWithin(self, node):
        return {'$geoWithin': GeoShapeParser().handle(self.get_arg(node, 0))}

class SymverFunc(Func):
    def handle_symver(self, node):
        return self.parse_arg(node, 0, SymverField())

class GenericFunc(StringFunc, IntFunc, ListFunc, DateTimeFunc,
                  IdFunc, EpochFunc, EpochUTCFunc, GeoFunc, SymverFunc):
    pass

#---Operators---#

class Operator(AstHandler):
    def __init__(self, field):
        self.field = field
    def handle_Eq(self, node):
        '''=='''
        return self.field.handle(node)
    def handle_NotEq(self, node):
        '''!='''
        return {'$ne': self.field.handle(node)}
    def handle_In(self, node):
        '''in'''
        try:
            elts = node.elts
        except AttributeError:
            raise ParseError('Invalid value type for `in` operator: {0}'.format(node.__class__.__name__),
                             col_offset=node.col_offset)
        return {'$in': list(map(self.field.handle, elts))}
    def handle_NotIn(self, node):
        '''not in'''
        try:
            elts = node.elts
        except AttributeError:
            raise ParseError('Invalid value type for `not in` operator: {0}'.format(node.__class__.__name__),
                             col_offset=node.col_offset)
        return {'$nin': list(map(self.field.handle, elts))}

class AlgebricOperator(Operator):
    def handle_Gt(self, node):
        '''>'''
        return {'$gt': self.field.handle(node)}
    def handle_Lt(self,node):
        '''<'''
        return {'$lt': self.field.handle(node)}
    def handle_GtE(self, node):
        '''>='''
        return {'$gte': self.field.handle(node)}
    def handle_LtE(self, node):
        '''<='''
        return {'$lte': self.field.handle(node)}

#---Field-Types---#

class Field(AstHandler):
    OP_CLASS = Operator

    SPECIAL_VALUES = {'None': None,
                      'null': None}

    def handle_Constant(self, node):
        try:
            return self.SPECIAL_VALUES[str(node.value)]
        except KeyError:
            raise ParseError('Invalid name: {0}'.format(node.value), node.col_offset, options=list(self.SPECIAL_VALUES))

    def handle_Name(self, node):
        try:
            return self.SPECIAL_VALUES[node.id]
        except KeyError:
            raise ParseError('Invalid name: {0}'.format(node.id), node.col_offset, options=list(self.SPECIAL_VALUES))

    def handle_operator_and_right(self, operator, right):
        return self.OP_CLASS(self).resolve(operator)(right)

class GeoField(Field):
    def handle_Call(self, node):
        return GeoFunc().handle(node)

class AlgebricField(Field):
    OP_CLASS = AlgebricOperator

class StringField(AlgebricField):
    def handle_Call(self, node):
        return StringFunc().handle(node)
    def handle_Str(self, node):
        return node.s
    def handle_Constant(self, node):
        return node.value

class IntField(AlgebricField):
    def handle_Constant(self, node):
        return node.value
    def handle_Num(self, node):
        return node.n
    def handle_UnaryOp(self, node):
        op_type = type(node.op)
        if (op_type == ast.USub):
            return - node.operand.value
        else:
            return node.operand.value

    def handle_Call(self, node):
        return IntFunc().handle(node)

class BoolField(Field):
    SPECIAL_VALUES = dict(Field.SPECIAL_VALUES,
                          **{'False': False,
                             'True': True,
                             'false': False,
                             'true': True})

class ListField(Field):
    def __init__(self, field=None):
        self._field = field
    def handle_List(self, node):
        return list(map((self._field or GenericField()).handle, node.elts))
    def handle_Call(self, node):
        return ListFunc().handle(node)

class DictField(Field):
    def __init__(self, field=None):
        self._field = field
    def handle_Dict(self, node):
        return dict((StringField().handle(key), (self._field or GenericField()).handle(value))
                    for key, value in zip(node.keys, node.values))

class DateTimeField(AlgebricField):
    def handle_Constant(self, node):
        return parse_date(node.value)
    def handle_Str(self, node):
        return parse_date(node)
    def handle_Num(self, node):
        return parse_date(node)
    def handle_Call(self, node):
        return DateTimeFunc().handle(node)

class EpochField(AlgebricField):
    def handle_Constant(self, node):
        return float(parse_date(node).strftime('%s.%f'))
    def handle_Str(self, node):
        return float(parse_date(node).strftime('%s.%f'))
    def handle_Num(self, node):
        return node.n
    def handle_Call(self, node):
        return EpochFunc().handle(node)

class EpochUTCField(AlgebricField):
    def handle_Constant(self, node):
        return timegm(parse_date(node).timetuple())
    def handle_Str(self, node):
        return timegm(parse_date(node).timetuple())
    def handle_Num(self, node):
        return node.n
    def handle_Call(self, node):
        return EpochUTCFunc().handle(node)

class IdField(AlgebricField):
    def handle_Constant(self, node):
        return bson.ObjectId(node.value)
    def handle_Str(self, node):
        return bson.ObjectId(node.s)
    def handle_Call(self, node):
        return IdFunc().handle(node)

class SymverField(AlgebricField):
    re_version = r'(\d+)\.(\d+)\.(\d+)\s*\((\d+)\)'

    def handle_Constant(self, node):
        d = re.findall(self.re_version, node.value)[0]
        return (int(d[0]) * 1e9) + (int(d[1]) * 1e6) + (int(d[2]) * 1e3) + int(d[3])

    def handle_Num(self, node):
        d = re.findall(self.re_version, node.value)[0]
        return (int(d[0]) * 1e9) + (int(d[1]) * 1e6) + (int(d[2]) * 1e3) + int(d[3])

    def handle_Call(self, node):
        return SymverFunc().handle(node)

class GenericField(IntField, BoolField, StringField, ListField, DictField, GeoField):
    def handle_Call(self, node):
        return GenericFunc().handle(node)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""V3DBHandler."""

from copy import deepcopy
from datetime import datetime
import hashlib
import importlib
import logging
import os

from migrate import create_column
import numpy as np
import pandas as pd
from sqlalchemy import sql, Column, Table, MetaData, inspect
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.types import VARCHAR, DECIMAL, BOOLEAN, TEXT
from sqlalchemy.dialects.mysql import DOUBLE
from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION
from tqdm import tqdm

from pydtk.db import V2BaseDBHandler as _V2BaseDBHandler
from pydtk.utils.utils import load_config
from pydtk.utils.utils import dicts_to_listed_dict_2d
from pydtk.utils.utils import dtype_string_to_dtype_object
from pydtk.utils.utils import serialize_dict_1d
from pydtk.utils.utils import deserialize_dict_1d

DB_HANDLERS = {}    # key: db_class, value: dict( key: db_engine, value: handler )


def map_dtype(dtype, db_engine='general'):
    """Mapper for dtype.

    Args:
        dtype (str): dtype in dataframe
        db_engine (str): DB engine

    Returns:
        (dict): { 'df': dtype for Pandas.DataFrame, 'sql': dtype for SQL table }

    """
    if dtype in ['int', 'int32', 'int64', 'float', 'float32', 'float64', 'double']:
        if db_engine in ['mysql', 'mariadb']:
            return {'df': 'double', 'sql': DOUBLE}
        elif db_engine in ['postgresql', 'timescaledb']:
            return {'df': 'double', 'sql': DOUBLE_PRECISION}
        else:
            return {'df': 'double', 'sql': DECIMAL}

    if dtype in ['bool']:
        return {'df': 'boolean', 'sql': BOOLEAN}

    if dtype in ['object', 'string', 'text']:
        return {'df': 'text', 'sql': TEXT}

    raise ValueError('Unsupported dtype: {}'.format(dtype))


def register_handlers():
    """Register handlers."""
    for filename in os.listdir(os.path.join(os.path.dirname(__file__))):
        if not os.path.isfile(os.path.join(os.path.dirname(__file__), filename)):
            continue
        if filename == '__init__.py':
            continue

        try:
            importlib.import_module(
                os.path.join('pydtk.db.v3.handlers',
                             os.path.splitext(filename)[0]).replace(os.sep, '.')
            )
        except ModuleNotFoundError:
            logging.debug('Failed to load handlers in {}'.format(filename))


def register_handler(db_classes, db_engines):
    """Register a DB-handler.

    Args:
        db_classes (list): list of db_class names (e.g. ['meta'])
        db_engines (list): list of supported db_engines (e.g. ['sqlite', 'mysql'])

    """
    def decorator(cls):
        for db_class in db_classes:
            if db_class not in DB_HANDLERS.keys():
                DB_HANDLERS.update({db_class: {}})
            for db_engine in db_engines:
                if db_engine not in DB_HANDLERS[db_class].keys():
                    DB_HANDLERS[db_class].update({db_engine: cls})
        return cls
    return decorator


class BaseDBHandler(_V2BaseDBHandler):
    """Base handler for db."""

    __version__ = 'v3'
    _df_class = 'base_df'

    def __new__(cls, db_class: str = None, db_engine: str = None, **kwargs) -> object:
        """Create object.

        Args:
            db_class (str): database class (e.g. 'meta')
            db_engine (str): database engine (e.g. 'sqlite')
            **kwargs: DB-handler specific arguments

        Returns:
            (object): the corresponding handler object

        """
        if cls is BaseDBHandler:
            handler = cls._get_handler(db_class, db_engine)
            return super(BaseDBHandler, cls).__new__(handler)
        else:
            return super(BaseDBHandler, cls).__new__(cls)

    @classmethod
    def _get_handler(cls, db_class, db_engine=None):
        """Returns an appropriate handler.

        Args:
            db_class (str): database class (e.g. 'meta')
            db_engine (str): database engine (e.g. 'sqlite')

        Returns:
            (handler): database handler object

        """
        # Load default config
        config = load_config(cls.__version__)

        # Check if the db_class is available
        if db_class not in DB_HANDLERS.keys():
            raise ValueError('Unsupported db_class: {}'.format(db_class))

        # Get db_engine from environment variable if not specified
        if db_engine is None:
            db_engine = os.environ.get('PYDTK_{}_DB_ENGINE'.format(db_class.upper()), None)

        # Get the default engine if not specified
        if db_engine is None:
            try:
                db_defaults = getattr(config.sql, db_class)
                db_engine = db_defaults.engine
            except (ValueError, AttributeError):
                raise ValueError('Could not find the default value')

        # Check if the corresponding handler is registered
        if db_engine not in DB_HANDLERS[db_class].keys():
            raise ValueError('Unsupported db_engine: {}'.format(db_engine))

        # Get a DB-handler supporting the engine
        return DB_HANDLERS[db_class][db_engine]

    def __init__(self, **kwargs):
        """Initialize BaseDBHandler.

        Args:
            **kwargs: kwargs

        """
        self._count_total = 0
        if 'db_class' in kwargs.keys():
            del kwargs['db_class']
        super(BaseDBHandler, self).__init__(**kwargs)

    def __next__(self):
        """Return the next item."""
        if self._cursor >= len(self.df):
            self._cursor = 0
            raise StopIteration()

        # Grab data
        data = self.df.take([self._cursor]).to_dict(orient='records')[0]

        # Delete internal column
        if 'uuid_in_df' in data.keys():
            del data['uuid_in_df']
        if 'creation_time_in_df' in data.keys():
            del data['creation_time_in_df']

        # Post-processes
        data = deserialize_dict_1d(data)

        # Increment
        self._cursor += 1

        return data

    def _initialize_df(self):
        """Initialize DF."""
        df = pd.concat(
            [pd.Series(name=c['name'],
                       dtype=dtype_string_to_dtype_object(c['dtype'])) for c in self.columns if c['name'] != 'uuid_in_df' and c['name'] != 'creation_time_in_df']  # noqa: E501
            + [pd.Series(name='uuid_in_df', dtype=str),
               pd.Series(name='creation_time_in_df', dtype=float)],
            axis=1
        )
        df.set_index('uuid_in_df', inplace=True)
        return df

    def _get_uuid_from_item(self, data_in):
        """Return UUID of the given item.

        Args:
            data_in (dict or pandas.Series): dict or Series containing data

        Returns:
            (str): UUID

        """
        hash_target_columns = self._config[self._df_class]['index_columns']

        item = data_in
        if isinstance(item, pd.Series):
            item = data_in.to_dict()

        pre_hash = ''.join([
            '{:.09f}'.format(item[column])
            if isinstance(item[column], float) else str(item[column])
            for column in hash_target_columns
            if column in item.keys()
        ])
        pre_hash = pre_hash.encode('utf-8')
        uuid = hashlib.md5(pre_hash).hexdigest()
        return uuid

    def _preprocess_list_of_dicts(self, data_in):
        """Preprocess list of dicts.

        Args:
            data_in (list): list of dicts containing data

        Returns:
            (dict): dict of lists = listed dict

        """
        data = deepcopy(data_in)

        # Serialize (convert list to str)
        self.logger.info('(Preprocess) Serializing...')
        for item in tqdm(data, desc='Serialization', leave=False):
            item = serialize_dict_1d(item)

            # Add df_uuid and creation_time_in_df
            item['uuid_in_df'] = self._get_uuid_from_item(item)
            item['creation_time_in_df'] = datetime.now().timestamp()

            # Add missing columns
            for column in self.columns:
                column_name, column_dtype = column['name'], column['dtype']
                if column_name not in item.keys():
                    item.update({column_name: None})
                else:
                    dtype_obj = dtype_string_to_dtype_object(column_dtype)
                    if dtype_obj is None:
                        continue
                    if item[column_name] is not None \
                       and not isinstance(item[column_name], dtype_obj):
                        try:
                            item[column_name] = dtype_obj(item[column_name])
                        except ValueError:
                            item[column_name] = np.nan

        # Convert dict to listed dict
        self.logger.info('(Preprocess) Converting...')
        data = dicts_to_listed_dict_2d(data)

        return data

    def _append_listed_dict_to_df(self, data, check_unique=False):
        """Append pre-processed dict to self._df.

        Args:
            data (dict): data to add
            check_unique (bool): if True, it will be checked that the data is unique in the db

        """
        self.df = pd.concat(
            [self.df, pd.DataFrame.from_dict(data).set_index('uuid_in_df')],
            sort=False
        )
        if check_unique:
            self.df = self.df[~self.df.index.duplicated(keep='last')]

    def _get_column_names_from_db(self):
        """Acquire one row from DB and get columns.

        Returns:
            (list): list for column names

        """
        q = 'select * from {0} limit 1'.format(self._quote(self.df_name))
        df = pd.read_sql_query(q, con=self._engine)
        columns = df.columns.tolist()
        return columns

    def read(self,
             df_name=None,
             query=None,
             where=None,
             group_by=None,
             order_by=None,
             limit=None,
             offset=None,
             disable_count_total=False,
             **kwargs):
        """Read data from SQL.

        Args:
            df_name (str): Dataframe name to read
            query (str SQL query or SQLAlchemy Selectable): query to select items
            where (str): query string for filtering items
            group_by (str): column name to group
            order_by (srt): column name to sort by
            limit (int): number of items to return per a page
            offset (int): offset of cursor
            disable_count_total (bool): if True, `self.count_total` will not be calculated
            **kwargs: kwargs for function `pandas.read_sql_query`
                      or `influxdb.DataFrameClient.query`

        """
        if df_name is not None:
            self.df_name = df_name

        # Create a query
        q = query
        q_count = 'select count(*) from ({}) as sub'.format(q) if q is not None else None
        if q is None:
            # Check if the table exits on DB
            if not inspect(self._engine).has_table(self.df_name):
                self.df = self._initialize_df()
                return

            # Create a query
            if self._config.current_db['engine'] in ['postgresql', 'timescaledb']:
                q = sql.select('*', from_obj=sql.table(self.df_name))
                if group_by is not None:
                    q = sql.select(
                        [sql.text(self._distinct(c))
                         for c in self._get_column_names_from_db()],
                        from_obj=sql.table(self.df_name))
                    q = q.group_by(sql.text(group_by))
                if where is not None:
                    q = q.where(sql.text(where))
                q_count = sql.select([sql.text('count(*)')], from_obj=sql.alias(q, 'sub'))
            else:
                q = sql.select('*', from_obj=sql.table(self.df_name))
                q_count = sql.select([sql.text('count(*)')], from_obj=sql.table(self.df_name))
                if group_by is not None:
                    q = sql.select(
                        [sql.text(self._distinct(c))
                         for c in self._get_column_names_from_db()],
                        from_obj=sql.table(self.df_name))
                    q = q.group_by(sql.text(group_by))
                    q_count = sql.select([sql.text('count(distinct {})'.format(group_by))],
                                         from_obj=sql.table(self.df_name))
                if where is not None:
                    q = q.where(sql.text(where))
                    q_count = q_count.where(sql.text(where))

            # Add common options
            if order_by is not None:
                q = q.order_by(sql.text(order_by))
            if limit is not None:
                q = q.limit(limit)
            if offset is not None:
                q = q.offset(offset)

        # Read table from DB
        try:
            # Calculate total number of rows
            if not disable_count_total:
                count = self._engine.execute(q_count).scalar()
            else:
                count = 0

            # Fetch data
            if 'index_col' not in kwargs.keys():
                kwargs.update({'index_col': 'uuid_in_df'})
            df = pd.read_sql_query(q, self._engine, **kwargs)
            df = df[~df.index.duplicated(keep='last')]
        except Exception as e:
            self.logger.warning('Could not execute SQL statement: "{0}" (reason: {1})'.
                                format(str(q), str(e)))
            count = 0
            df = self._initialize_df()

        self._count_total = count
        self.df = df

    def save(self, df=None, remove_duplicates=False, **kwargs):
        """Save data to SQL.

        Args:
            df (pandas.DataFrame): DataFrame to save (if None, self.df will be saved)
            remove_duplicates (bool): if True, duplicated rows will be removed
            **kwargs: kwargs for function `pandas.dataframe.to_sql`

        """
        dataframe = df.copy() if df is not None else self.df.copy()
        dataframe = dataframe[~dataframe.index.duplicated(keep='last')]

        if 'index' not in kwargs.keys():
            kwargs.update({'index': True})
        self._prepare_columns()
        dataframe.to_sql(self.df_name,
                         self._engine,
                         if_exists='append',
                         dtype={'uuid_in_df': VARCHAR(32)},
                         **kwargs)

        if remove_duplicates:
            self._remove_duplicates()

    def _prepare_columns(self):
        try:
            existing_columns = self._get_column_names_from_db()
        except (OperationalError, ProgrammingError):
            return

        # Get table
        meta = MetaData(bind=self._engine)
        table = Table(self.df_name, meta, autoload=True)

        # Add columns
        for column_name in set([c['name'] for c in self.columns]).difference(existing_columns):
            column = next((filter(lambda c: c['name'] == column_name, self.columns)))
            column_name = column['name']
            if column_name in ['uuid_in_df']:
                column_dtype = VARCHAR(32)
            elif column_name in ['creation_time_in_df']:
                column_dtype = map_dtype('double', self._config.current_db['engine'])['sql']
            else:
                column_dtype = map_dtype(column['dtype'], self._config.current_db['engine'])['sql']
            create_column(Column(column_name, column_dtype), table)

    def _quote(self, value):
        if self._config.current_db['engine'] in ['mariadb']:
            return '`{}`'.format(value)
        else:
            return '"{}"'.format(value)

    def _distinct(self, column_name):
        """Return selection key to get the distinct element.

        Args:
            column_name (str): name of the column

        Returns:
            (str): selection key

        """
        column_dtype = None
        if column_name in self._get_column_names():
            column_info = next(iter(filter(lambda c: c['name'] == column_name, self.columns)))
            column_dtype = column_info['dtype']

        if self._config.current_db['engine'] in ['postgresql', 'timescaledb']:
            if column_dtype in ['string[]']:
                return 'string_agg({0}, \';\') as {0}'.format(column_name)
            return 'min({0}) as {0}'.format(column_name)
        else:
            if column_dtype in ['string[]']:
                return 'group_concat({0}, \';\') as {0}'.format(column_name)
            return column_name

    def _remove_duplicates(self):
        temp_table_name = self.df_name + '_' + datetime.now().strftime('%s')

        columns = self._get_column_names_from_db()

        q = 'create table {0} as ' \
            'select {2} from {1} ' \
            'where creation_time_in_df in (' \
            '  select max(creation_time_in_df) from {1}' \
            '  group by uuid_in_df' \
            ') ' \
            'group by uuid_in_df'. \
            format(self._quote(temp_table_name),
                   self._quote(self.df_name),
                   ','.join([self._distinct(self._quote(c)) for c in columns]))
        self._engine.execute(q)

        # Create index
        if self._config.current_db['engine'] in ['sqlite']:
            q = 'create index index_uuid_in_df on {} (uuid_in_df)'.format(
                self._quote(self.df_name)
            )
        elif self._config.current_db['engine'] in ['mysql', 'mariadb']:
            q = 'alter table {} add index index_uuid_in_df(uuid_in_df)'.format(
                self._quote(self.df_name)
            )
        elif self._config.current_db['engine'] in ['postgresql', 'timescaledb']:
            q = 'create index on {} (uuid_in_df)'.format(self._quote(self.df_name))
        else:
            raise ValueError('Unsupported engine: {}'.format(
                self._config.current_db['engine']
            ))
        self._engine.execute(q)

        # Drop deprecated table if exists
        if inspect(self._engine).has_table(self.df_name + '_deprecated'):
            self._engine.execute('drop table {0}'.
                                 format(self._quote(self.df_name + '_deprecated')))

        # Deprecate the original table
        self._engine.execute('alter table {0} rename to {1}'.
                             format(self._quote(self.df_name),
                                    self._quote(self.df_name + '_deprecated')))

        # Rename the temporal table
        self._engine.execute('alter table {0} rename to {1}'.
                             format(self._quote(temp_table_name), self._quote(self.df_name)))

        # Drop the deprecated table and temporal table
        self._engine.execute('drop table {0}'.format(self._quote(self.df_name + '_deprecated')))

    @property
    def columns(self):
        """Return columns of DF."""
        if self._columns is not None:
            return self._columns
        try:
            return self._config[self._df_class]['columns']
        except KeyError:
            return []

    @columns.setter
    def columns(self, value):
        """Set self.columns."""
        if not isinstance(value, list):
            raise ValueError('Columns must be a list of dicts with keys "name" and "dtype".')
        if len(value) < 1:
            raise ValueError('At least one item must be in the list.')
        if not isinstance(value[0], dict) or "name" not in value[0].keys() or "dtype" not in value[0].keys():
            raise ValueError('Columns must be a list of dicts with keys "name" and "dtype".')

        try:
            _ = pd.concat(
                [pd.Series(name=c['name'],
                           dtype=dtype_string_to_dtype_object(c['dtype'])) for c in value]
                + [pd.Series(name='uuid_in_df', dtype=str),
                   pd.Series(name='creation_time_in_df', dtype=float)],
                axis=1
            )
        except KeyError as e:
            raise ValueError('Unrecognized value: {}'.format(str(e)))

        self._columns = value

    @property
    def df(self):
        """Return df."""
        return self._df

    @df.setter
    def df(self, value):
        """Setter for self.df."""
        if not isinstance(value, pd.DataFrame):
            raise ValueError('Only pandas dataframe is accepted.')

        if len(value) > 0:
            # Set columns based on the given DF
            self.columns = [
                {'name': c,
                 'dtype': map_dtype(dtype.name, self._config.current_db['engine'])['df']}
                for c, dtype in value.dtypes.to_dict().items()
            ]

            # Add column 'uuid_in_df' and 'creation_time_in_df'
            if 'uuid_in_df' not in self._get_column_names():
                self.columns += [{'name': 'uuid_in_df', 'dtype': 'str'}]
            if 'uuid_in_df' not in value.columns.to_list():
                value['uuid_in_df'] = value.apply(lambda x: self._get_uuid_from_item(x), axis=1)
            if 'creation_time_in_df' not in self._get_column_names():
                self.columns += [{'name': 'creation_time_in_df', 'dtype': 'double'}]
            if 'creation_time_in_df' not in value.columns.to_list():
                value['creation_time_in_df'] = datetime.now().timestamp()
            value.set_index('uuid_in_df', inplace=True)

        self._df = value

    @property
    def count_total(self):
        """Return total number of rows."""
        return self._count_total


register_handlers()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""Search engine for statistics."""

from .. import TimeSeriesDBHandler as _TimeSeriesDBHandler
from . import BaseDBSearchEngine as _BaseDBSearchEngine


class TimeSeriesDBSearchEngine(_BaseDBSearchEngine):
    """Search engine for TimeSeriesDB."""

    def __init__(self,
                 db_handler: _TimeSeriesDBHandler,
                 content: str,
                 span: float):
        """Initialize StatisticsDBSearchEngine.

        Args:
            db_handler (V2TimeSeriesDBHandler): database handler
            content (str): target content (e.g. /vehicle/acceleration)
            span (float): span of statistics in seconds (e.g. 60)

        """
        super(TimeSeriesDBSearchEngine, self).__init__(db_handler)
        self._content = content
        self._span = span

    def add_condition(self, condition: str):
        """Add a search condition.

        Args:
            condition (str): search condition
                             (e.g. '"/vehicle/acceleration/accel_linear_x/mean" > 0')

        """
        if '"' not in condition:
            raise ValueError('column name must be quoted '
                             '(e.g. "/vehicle/acceleration/accel_linear_x/mean" > 0)')
        self._conditions.append(condition)

    @property
    def select(self):
        """Return select query.

        Returns:
            (str): query

        """
        table_name = self._db_handler._config.statistics_df['name']\
            .replace('/', '\/')\
            .replace('{span}', '{0:.01f}'.format(self._span).replace('.', '\.'))\
            .replace('{record_id}', '.*')\
            .replace('{content}', self._content.replace('/', '\/'))
        return 'select * from /{}/'.format(table_name)



#!/usr/bin/env python3

# Copyright Toolkit Authors

import importlib
import json
import os

import pprint


class Model(object):
    """Model related operations."""

    @staticmethod
    def list(
        **kwargs
    ):
        """List models."""
        from pydtk.models import MODELS_BY_PRIORITY
        print('Available models with priorities:')
        pprint.pprint(MODELS_BY_PRIORITY)

    @staticmethod
    def is_available(
        file: str,
        content: str = None,
    ):
        """Test if available models exist against the given file.

        Args:
            file (str): Path to a file
            content (str): Content in the file to read

        """
        from pydtk.io import BaseFileReader
        from pydtk.io import NoModelMatchedError
        reader = BaseFileReader()
        try:
            _ = reader.read(path=file, contents=content, as_ndarray=False)
            print('True')
        except NoModelMatchedError:
            print('False')

    @staticmethod
    def generate(
        target: str,
        model: str = None,
        from_file: str = None,
        template: str = None,
        database_id: str = 'default',
        record_id: str = None,
        content: str = 'content',
        base_dir: str = None,
    ):
        """Generate template or metadata from a model or a file.

        Args:
            target (str): What to to generate ('template' or 'metadata')
            model (str): Model to use (e.g., 'pydtk.models.csv.GenericCsvModel')
            from_file (str): File path to generate metadata from
            template (str): JSON file to use as a template of metadata
            database_id (str): Database ID
            record_id (str): Record ID
            content (str): Content (key of dict `contents`)
            base_dir (str): Base directory

        """
        assert target in ['template', 'metadata'], 'Target must be either "template" or "metadata"'

        from pydtk.db import DBHandler
        from pydtk.io import BaseFileReader
        from pydtk.models import MetaDataModel
        from pydtk.utils.utils import load_config

        default_config = load_config('v4').bin.make_meta
        data = {k: None for k in default_config['common_item'].keys()}

        db_handler = DBHandler(db_class='meta', database_id=database_id)
        config = db_handler.config

        if target == 'template':
            data.update({
                column['name']: None
                for column in
                config['columns']
            })

            if template:
                data.update(template)

        elif target == 'metadata':
            if template:
                f = open(template, 'r')
                template_data = f.read()
                f.close()
                data.update(json.loads(template_data))

            if from_file is None:
                raise ValueError(
                    'Please specify a file to generate metadata from using option "--from-file"'
                )

            path = os.path.abspath(from_file)
            if base_dir:
                path = os.path.relpath(from_file, base_dir)

            data["path"] = path

            # Get model object
            if model is None:
                metadata = MetaDataModel(data={**data, 'path': from_file})
                model = BaseFileReader._select_model(metadata)
            else:
                model = getattr(importlib.import_module('.'.join(model.split('.')[:-1])),
                                model.split('.')[-1])

            # Get contents and timestamps
            try:
                data["contents"] = model.generate_contents_meta(path=from_file, content_key=content)
            except NotImplementedError:
                pass
            try:
                data["start_timestamp"], data["end_timestamp"] = \
                    model.generate_timestamp_meta(path=from_file)
            except NotImplementedError:
                pass

        # Post process
        if record_id is not None:
            data['record_id'] = record_id

        # Display
        print(json.dumps(data, indent=4, default=_default_json_handler))


def _default_json_handler(o):
    return o.__str__()



from setuptools import setup

__version__ = '0.5.0'

setup(name='pql',
      version=__version__,
      description='A python expression to MongoDB query translator',
      author='Alon Horev',
      author_email='alon@horev.net',
      url='https://github.com/alonho/pql',
      classifiers = ["Development Status :: 3 - Alpha",
                     "Intended Audience :: Developers",
                     "Programming Language :: Python :: 3.3",
                     "Programming Language :: Python :: 3.5",
                     "Operating System :: POSIX :: Linux",
                     "Operating System :: MacOS :: MacOS X"],
      license='BSD',
      # I know it's bad practice to not specify a pymongo version, but we only
      # require the bson.ObjectId type, It's safe to assume it won't change (famous last words)
      install_requires=['pymongo',
                        'python-dateutil'],
      packages=['pql'])


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""V4DBHandler."""

from collections import MutableMapping
from copy import deepcopy
from datetime import datetime
import hashlib
import importlib
import inspect
import logging
import os

from attrdict import AttrDict
from deepmerge import Merger
import pandas as pd

from pydtk.utils.utils import \
    load_config, \
    dtype_string_to_dtype_object, \
    _deepmerge_append_list_unique
from pydtk.db.exceptions import DatabaseNotInitializedError, InvalidDatabaseConfigError
from pydtk.db.v4.engines import DB_ENGINES

DB_HANDLERS = {}  # key: db_class, value: dict( key: db_engine, value: handler )

logger = logging.getLogger(__name__)


def register_handlers():
    """Register handlers."""
    for filename in os.listdir(os.path.join(os.path.dirname(__file__))):
        if not os.path.isfile(os.path.join(os.path.dirname(__file__), filename)):
            continue
        if filename == '__init__.py':
            continue

        try:
            importlib.import_module(
                os.path.join(
                    "pydtk.db.v4.handlers", str(os.path.splitext(filename)[0])
                ).replace(os.sep, ".")
            )
        except ModuleNotFoundError:
            logger.warning('Failed to load handlers in {}'.format(filename))


def register_handler(db_classes, db_engines):
    """Register a DB-handler.

    Args:
        db_classes (list): list of db_class names (e.g. ['meta'])
        db_engines (list): list of supported db_engines (e.g. ['tinydb', 'mongodb'])

    """

    def decorator(cls):
        for db_class in db_classes:
            if db_class not in DB_HANDLERS.keys():
                DB_HANDLERS.update({db_class: {}})
            for db_engine in db_engines:
                if db_engine not in DB_HANDLERS[db_class].keys():
                    DB_HANDLERS[db_class].update({db_engine: cls})
        return cls

    return decorator


def _fix_data_type(data_in, columns, inplace=False, aggregated=False):
    """Fix dtype of the input data.

    Args:
        data_in (dict): input-data
        columns (list): column configurations
        inplace (bool): if True, this function modifies the input argument `data_in`
        aggregated (bool): whether the data is aggregated or not

    Returns:
        (dict): data with corrected dtypes

    """
    data = data_in if inplace else deepcopy(data_in)

    def _fix_dtype(_data, column_conf):
        if column_conf['dtype'].lower() in ['string', 'str']:
            return str(_data)
        elif column_conf['dtype'].lower() in ['integer', 'int']:
            return int(_data)
        elif column_conf['dtype'].lower() in ['float', 'double', 'number']:
            return float(_data)
        elif column_conf['dtype'].lower() in ['list'] or column_conf['dtype'].endswith('[]'):
            return list(_data)
        elif column_conf['dtype'].lower() in ['dict']:
            return dict(_data)
        elif column_conf['dtype'].lower() in ['datetime']:
            if isinstance(_data, str):
                if ':' in _data or '-' in _data:
                    # ISO format
                    return datetime.fromisoformat(_data)
                else:
                    raise ValueError(
                        f'Unknown format of datetime: "{_data}"'
                        f'Please make sure to use ISO format (YYYY-mm-dd HH:MM:SS.ffffff)'
                    )
            elif isinstance(_data, float):
                # Epoch time
                return datetime.fromtimestamp(_data)
            elif isinstance(_data, datetime):
                return _data
            else:
                raise TypeError(
                    f'Unsupported type ({type(_data).__name__}): "{_data}"'
                    f'Please make sure the type of value "{_data}" is either string or float'
                )
        else:
            pass

        return _data

    for key in data.keys():
        if key.startswith('_'):
            continue
        try:
            column_conf = next(filter(lambda c: c['name'] == key, columns))
        except StopIteration:
            raise InvalidDatabaseConfigError(f'Column "{key}" not found')
        if 'dtype' not in column_conf.keys():
            continue
        if data[key] is None:
            continue
        if not aggregated:
            data[key] = _fix_dtype(data[key], column_conf)
        else:
            if 'aggregation' not in column_conf.keys():
                # First
                data[key] = _fix_dtype(data[key], column_conf)
            if column_conf['aggregation'].lower() in ['first', 'min', 'max']:
                data[key] = _fix_dtype(data[key], column_conf)
            elif column_conf['aggregation'].lower() in ['push']:
                data[key] = [_fix_dtype(_data, column_conf) for _data in data[key]]
            else:
                pass

    return data


class BaseDBHandler(object):
    """Base handler for db."""

    __version__ = 'v4'
    db_defaults = load_config(__version__).db.connection.base
    _df_class = 'base_df'
    _config = AttrDict()
    _df_name = 'base_df'
    _columns = None
    _read_conditions = {}

    def __new__(cls, db_class: str = None, db_engine: str = None, **kwargs) -> object:
        """Create object.

        Args:
            db_class (str): database class (e.g. 'meta')
            db_engine (str): database engine (e.g. 'sqlite')
            **kwargs: DB-handler specific arguments

        Returns:
            (object): the corresponding handler object

        """
        if cls is BaseDBHandler:
            handler = cls._get_handler(db_class, db_engine)
            return super(BaseDBHandler, cls).__new__(handler)
        else:
            return super(BaseDBHandler, cls).__new__(cls)

    @classmethod
    def _get_handler(cls, db_class, db_engine=None):
        """Returns an appropriate handler.

        Args:
            db_class (str): database class (e.g. 'meta')
            db_engine (str): database engine (e.g. 'tinydb')

        Returns:
            (handler): database handler object

        """
        # Load default config
        config = load_config(cls.__version__)

        # Check if the db_class is available
        if db_class not in DB_HANDLERS.keys():
            raise ValueError('Unsupported db_class: {}'.format(db_class))

        # Get db_engine from environment variable if not specified
        if db_engine is None:
            db_engine = os.environ.get('PYDTK_{}_DB_ENGINE'.format(db_class.upper()), None)

        # Get the default engine if not specified
        if db_engine is None:
            try:
                db_defaults = getattr(config.db.connection, db_class)
                db_engine = db_defaults.engine
            except (ValueError, AttributeError):
                raise ValueError('Could not find the default value for `db_engine`')

        # Check if the corresponding handler is registered
        if db_engine not in DB_HANDLERS[db_class].keys():
            raise ValueError('Unsupported db_engine: {}'.format(db_engine))

        # Get a DB-handler supporting the engine
        return DB_HANDLERS[db_class][db_engine]

    def __init__(self,
                 db_engine=None,
                 db_host=None,
                 db_name=None,
                 db_username=None,
                 db_password=None,
                 df_name=None,
                 read_on_init=False,
                 **kwargs):
        """Initialize BaseDBHandler.

        Args:
            db_engine (str): database engine (if None, the one in the config file will be used)
            db_host (str): database HOST (if None, the one in the config file will be used)
            db_name (str): database name (if None, the one in the config file will be used)
            db_username (str): username (if None, the one in the config file will be used)
            db_password (str): password (if None, the one in the config file will be used)
            df_name (str): dataframe (table in DB) name (if None, class default value will be used)
            read_on_init (bool): if True, dataframe will be read from database on initialization

        """
        super(BaseDBHandler, self).__init__()
        self.logger = logging.getLogger(__name__)
        self._cursor = 0
        self._data = {}
        self._uuids_duplicated = []
        self._uuids_to_remove = []
        self._tables_to_drop = []
        self._count_total = 0
        if df_name is not None:
            self.df_name = df_name

        # Load config
        config = load_config(self.__version__)
        try:
            self._config = ConfigDict(config['db']['df_class'][self._df_class])
        except KeyError:
            self._config = ConfigDict()

        # Initialize deepmerger
        self._merger = Merger(
            # pass in a list of tuple, with the
            # strategies you are looking to apply
            # to each type.
            [(dict, ['merge']),
             (list, [_deepmerge_append_list_unique, 'append'])],
            # next, choose the fallback strategies,
            # applied to all other types:
            ["override"],
            # finally, choose the strategies in
            # the case where the types conflict:
            ["override"]
        )

        # Initialize database
        self._initialize_engine(db_engine, db_host, db_name, db_username, db_password)
        self._load_config_from_db()

        # Fetch table
        if read_on_init:
            self.read()

    def __len__(self):
        """Return number of records."""
        return len(self.data)

    def __iter__(self):
        """Return iterator."""
        return self

    def __next__(self):
        """Return the next item."""
        if self._cursor >= len(self):
            self._cursor = 0
            raise StopIteration()

        # Grab data
        data = self[self._cursor]

        # Increment
        self._cursor += 1

        return data

    def __getitem__(self, idx, remove_internal_columns=True):
        """Return data at index idx.

        Args:
            idx (int): index of the target data
            remove_internal_columns (bool): if True, internal columns are removed

        Returns:
            (dict): data

        """
        data = self.data[idx]

        # Delete internal column
        if remove_internal_columns:
            if '_uuid' in data.keys():
                del data['_uuid']
            if '_creation_time' in data.keys():
                del data['_creation_time']

        return data

    def _initialize_engine(self,
                           db_engine=None,
                           db_host=None,
                           db_name=None,
                           db_username=None,
                           db_password=None):
        """Initialize DB engine.

        Args:
            db_engine (str): database engine (if None, the one in the config file will be used)
            db_host (str): database HOST (if None, the one in the config file will be used)
            db_name (str): database name (if None, the one in the config file will be used)
            db_username (str): username (if None, the one in the config file will be used)
            db_password (str): password (if None, the one in the config file will be used)

        """
        if db_engine is None:
            db_engine = self.db_defaults.engine
        if db_host is None:
            db_host = self.db_defaults.host
            if db_username is None:
                db_username = self.db_defaults.username
            if db_password is None:
                db_password = self.db_defaults.password

        self._db_engine = db_engine

        if db_engine in DB_ENGINES.keys():
            self._db = DB_ENGINES[db_engine].connect(
                db_host=db_host,
                db_name=db_name,
                db_username=db_username,
                db_password=db_password,
                collection_name=self._df_name,
                handler=self,
            )
            self._config_db = DB_ENGINES[db_engine].connect(
                db_host=db_host,
                db_name=db_name,
                db_username=db_username,
                db_password=db_password,
                collection_name='--config--{}'.format(self._df_name),
                handler=self,
            )
        else:
            raise ValueError("Unsupported engine: {}".format(db_engine))

    def _load_config_from_db(self):
        """Load configs from DB."""
        if self._db_engine not in DB_ENGINES.keys():
            return
        try:
            candidates = DB_ENGINES[self._db_engine].read(self._config_db, handler=self)
            if len(candidates) > 0:
                if isinstance(candidates[0][0], dict):
                    self._config = ConfigDict(candidates[0][0])
                else:
                    raise TypeError('Unexpected type')
        except Exception as e:
            self.logger.warning('Failed to load configs from DB: {}'.format(str(e)))

    def _save_config_to_db(self):
        """Save configs to DB."""
        if self._db_engine not in DB_ENGINES.keys():
            return
        try:
            config = dict(self._config)
            config.update({'_uuid': '__config__'})
            config = [config]
            DB_ENGINES[self._db_engine].upsert(self._config_db, data=config, handler=self)
        except Exception as e:
            self.logger.warning('Failed to save configs to DB: {}'.format(str(e)))

    def _get_uuid_from_item(self, data_in):
        """Return UUID of the given item.

        Args:
            data_in (dict or pandas.Series): dict or Series containing data

        Returns:
            (str): UUID

        """
        hash_target_columns = \
            self._config['index_columns'] if 'index_columns' in self._config.keys() else []

        item = data_in
        if isinstance(item, pd.Series):
            item = data_in.to_dict()

        pre_hash = ''.join([
            '{:.09f}'.format(item[column])
            if isinstance(item[column], float) else
            str(item[column].keys()) if isinstance(item[column], dict) else
            str(item[column])
            for column in hash_target_columns
            if column in item.keys()
        ])
        pre_hash = pre_hash.encode('utf-8')
        uuid = hashlib.md5(pre_hash).hexdigest()
        return uuid

    def _read(self, **kwargs):
        if self._db_engine is None:
            raise DatabaseNotInitializedError()
        elif self._db_engine in DB_ENGINES.keys():
            func = DB_ENGINES[self._db_engine].read
            available_args = set(inspect.signature(func).parameters.keys())
            unavailable_args = set([k for k, v in kwargs.items() if v is not None]) \
                .difference(available_args)
            if len(unavailable_args) > 0:
                self.logger.warning('DB-engine "{0}" does not support args: {1}'.
                                    format(self._db_engine, list(unavailable_args)))
            return func(self._db, handler=self, **kwargs)
        else:
            raise ValueError('Unsupported DB engine: {}'.format(self._db_engine))

    def read(self,
             df_name=None,
             query=None,
             pql=None,
             where=None,
             group_by=None,
             order_by=None,
             limit=None,
             offset=None,
             **kwargs):
        """Read data from SQL.

        Args:
            df_name (str): Deprecated. Dataframe name to read
            query (str SQL query or SQLAlchemy Selectable): query to select items
            pql (PQL): Python-Query-Language to select items
            where (str): query string for filtering items
            group_by (str): column name to group
            order_by (srt): column name to sort by
            limit (int): number of items to return per a page
            offset (int): offset of cursor
            **kwargs: kwargs for function `pandas.read_sql_query`
                      or `influxdb.DataFrameClient.query`

        """
        if df_name is not None:
            logger.warning('Specifying `df_name` in read() is deprecated '
                           'and will be unsupported in the future release as '
                           'it may cause unexpected behavior')
            self.df_name = df_name

        # load config from DB
        self._load_config_from_db()

        # query data
        data, self._count_total = self._read(
            query=query,
            pql=pql,
            where=where,
            group_by=group_by,
            order_by=order_by,
            limit=limit,
            offset=offset,
            **kwargs
        )

        # Check if UUID exists in each value
        for value in data:
            if '_uuid' not in value.keys():
                raise ValueError('"_uuid" not found in data')

        # Store conditions
        self._read_conditions = {
            'query': query,
            'pql': pql,
            'where': where,
            'group_by': group_by,
            'order_by': order_by,
            'limit': limit,
            'offset': offset,
            **kwargs
        }

        # Fix data-type
        columns = self._config['columns'] if 'columns' in self._config.keys() else []
        for i in range(len(data)):
            _fix_data_type(data[i], columns, inplace=True, aggregated=group_by is not None)

        self._data = {record['_uuid']: record for record in data}
        self._uuids_duplicated = []

    def _upsert(self, data):
        """Upsert data to DB.

        Args:
            data (list): data to save

        """
        if self._db_engine is None:
            raise DatabaseNotInitializedError()
        elif self._db_engine in DB_ENGINES.keys():
            DB_ENGINES[self._db_engine].upsert(self._db, data)
        else:
            raise ValueError('Unsupported DB engine: {}'.format(self._db_engine))

    def save(self, **kwargs):
        """Save data to DB."""
        self._remove(self._uuids_to_remove)
        for table_name in self._tables_to_drop:
            self._drop_table(table_name)
        self._uuids_to_remove = []
        self._uuids_duplicated = []
        self._upsert(list(self._data.values()))
        self._save_config_to_db()

    def _remove(self, uuids):
        """Remove data from DB.

        Args:
            uuids [str]: A list of unique IDs

        """
        if self._db_engine is None:
            raise DatabaseNotInitializedError()
        elif self._db_engine in DB_ENGINES.keys():
            DB_ENGINES[self._db_engine].remove(self._db, uuids)
        else:
            raise ValueError('Unsupported DB engine: {}'.format(self._db_engine))

    def _drop_table(self, name):
        """Drop table from DB.

        Args:
            name (str): Name of table (collection)

        """
        if self._db_engine is None:
            raise DatabaseNotInitializedError()
        elif self._db_engine in DB_ENGINES.keys():
            DB_ENGINES[self._db_engine].drop_table(self._db, name)
        else:
            raise ValueError('Unsupported DB engine: {}'.format(self._db_engine))

    def drop_table(self, name):
        """Drop table from DB (This will no be applied unless `save()` is called).

        Args:
            name (str): Name of table (collection)

        """
        # Remove from DB on saving
        self._tables_to_drop.append(name)

        # Remove the corresponding config if exist
        if self._db_engine is None:
            raise DatabaseNotInitializedError()
        elif self._db_engine in DB_ENGINES.keys():
            if DB_ENGINES[self._db_engine].exist_table(self._db, '--config--{}'.format(name)):
                self._tables_to_drop.append('--config--{}'.format(name))
        else:
            raise ValueError('Unsupported DB engine: {}'.format(self._db_engine))

    def add_data(self, data_in, strategy='overwrite', ignore_dtype_mismatch=False, **kwargs):
        """Add data to DB-handler.

        Args:
            data_in (dict): a dict containing data
            strategy (str): 'merge' or 'overwrite'
            ignore_dtype_mismatch (bool): if True, data type will not be modified
                                          regardless of column specifications

        """
        assert strategy in ['merge', 'overwrite'], 'Unknown strategy.'

        data = deepcopy(data_in)
        if '_uuid' not in data.keys() or '_creation_time' not in data.keys():
            # Add _uuid and _creation_time
            data['_uuid'] = self._get_uuid_from_item(data)
            data['_creation_time'] = datetime.now().timestamp()

        if data['_uuid'] in self._data.keys():
            if strategy == 'merge':
                base_data = self._data[data['_uuid']]
                data = self._merger.merge(base_data, data)
            self._uuids_duplicated += [data['_uuid']]

        # Add new columns (keys) to config
        columns = self._config['columns'] if 'columns' in self._config.keys() else []
        columns_existing = [c['name'] for c in columns]
        columns_in_data = list(data_in.keys())
        new_columns = []
        for new_column in set(columns_in_data).difference(columns_existing):
            name = new_column
            dtype = type(data_in[new_column]).__name__
            aggregation = 'first'
            new_columns.append({
                'name': name,
                'dtype': dtype,
                'aggregation': aggregation,
                'display_name': name,
            })
        columns += new_columns

        # Fix dtype of the input data
        if not ignore_dtype_mismatch:
            _fix_data_type(data, columns, inplace=True)

        # Update self
        self._config['columns'] = columns
        self._data.update({data['_uuid']: data})

    def remove_data(self, data):
        """Remove data-record from DB-handler.

        Args:
            data (dict): a dict containing the target data or '_uuid' in keys.

        """
        if '_uuid' in data.keys():
            uuid = data['_uuid']
        else:
            uuid = self._get_uuid_from_item(data)

        # Remove from in-memory data
        if uuid in self._data.keys():
            del self._data[uuid]

        # Remove from DB on saving
        self._uuids_to_remove.append(uuid)

    def _df_from_dicts(self, dicts):
        """Create a DataFrame from a list of dicts.

        Args:
            dicts (list): list of dicts

        Returns:
            (pd.DataFrame): a data-frame

        """
        columns = self._config['columns'] if 'columns' in self._config.keys() else []
        df = pd.concat(
            [pd.Series(name=c['name'],
                       dtype=dtype_string_to_dtype_object(c['dtype']))
             for c in columns
             if c['name'] != '_uuid' and c['name'] != '_creation_time']  # noqa: E501
            + [pd.Series(name='_uuid', dtype=str),
               pd.Series(name='_creation_time', dtype=float)],
            axis=1
        )
        df.set_index('_uuid', inplace=True)
        df = pd.concat([df, pd.DataFrame.from_records(dicts)])

        # Apply offset to index
        if 'offset' in self._read_conditions.keys() and self._read_conditions['offset'] is not None:
            df.index += self._read_conditions['offset']

        return df

    def _to_display_names(self, df, inplace=False):
        """Rename columns to display-names.

        Args:
            df (pd.DataFrame): A data-frame
            inplace (bool): If true, df will be modified in-place.

        Returns:
            (pd.DataFrame): A data-frame with renamed columns if inplace=false, otherwise None.

        """
        column_renames = {
            c['name']: c['display_name'] for c in self._config['columns']
            if isinstance(c, dict) and 'name' in c.keys() and 'display_name' in c.keys()
        }
        return df.rename(columns=column_renames, inplace=inplace)

    @property
    def data(self):
        """Return data.

        Returns:
            (list): list of dicts

        """
        return list(self._data.values())

    @data.setter
    def data(self, data):
        """Setter for self.data.

        Args:
            data (list): new-data

        """
        assert isinstance(data, list)
        for value in data:
            self.add_data(value)

    @property
    def columns(self):
        """Return columns of DF."""
        df = self._df_from_dicts(self.data)
        return df.columns.tolist()

    @property
    def df(self):
        """Return df."""
        df = self._df_from_dicts(self.data)
        df = self._to_display_names(df)
        return df

    @property
    def count_total(self):
        """Return total number of rows."""
        return self._count_total

    @property
    def config(self):
        """Return config."""
        return self._config


class ConfigDict(MutableMapping, dict):
    """A dictionary class customized for configs of DB-handlers."""

    def __getitem__(self, key):
        """Getter."""
        return dict.__getitem__(self, key)

    def __setitem__(self, key, value, force=False):
        """Setter."""
        if key.startswith('_') and not force:
            raise KeyError('key "{}" is not editable'.format(key))
        dict.__setitem__(self, key, value)

    def __delitem__(self, key):
        """Delete item."""
        dict.__delitem__(self, key)

    def __iter__(self):
        """Iterator."""
        return dict.__iter__(self)

    def __len__(self):
        """Length."""
        return dict.__len__(self)

    def __contains__(self, x):
        """Check if self contains the key."""
        return dict.__contains__(self, x)


register_handlers()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""V2DBHandler."""
_extra_supports = {
    'influxdb': True,
}

from copy import deepcopy
from datetime import datetime
import hashlib
import logging

from attrdict import AttrDict
import numpy as np
try:
    from influxdb import DataFrameClient
except ImportError:
    _extra_supports['influxdb'] = False
import pandas as pd
import sqlalchemy
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy import sql
from tqdm import tqdm

from pydtk.utils.utils import load_config
from pydtk.utils.utils import dicts_to_listed_dict_2d
from pydtk.utils.utils import dtype_string_to_dtype_object
from pydtk.utils.utils import serialize_dict_1d
from pydtk.utils.utils import deserialize_dict_1d


class BaseDBHandler(object):
    """Handler for db."""

    __version__ = 'v2'
    db_defaults = load_config(__version__).sql.base
    _config = AttrDict()
    _df_name = 'base_df'
    _columns = None

    @classmethod
    def default_config(cls):
        """Return default configurations.

        Returns:
            (AttrDict): default configurations

        """
        return load_config(cls.__version__)

    def __init__(self,
                 db_engine=None,
                 db_host=None,
                 db_name=None,
                 db_username=None,
                 db_password=None,
                 df_name=None,
                 read_on_init=True):
        """Initialize BaseDBHandler.

        Args:
            db_engine (str): database engine (if None, the one in the config file will be used)
            db_host (str): database HOST (if None, the one in the config file will be used)
            db_name (str): database name (if None, the one in the config file will be used)
            db_username (str): username (if None, the one in the config file will be used)
            db_password (str): password (if None, the one in the config file will be used)
            df_name (str): dataframe (table in DB) name (if None, class default value will be used)
            read_on_init (bool): if True, dataframe will be read from database on initialization

        """
        self._cursor = 0
        self.logger = logging.getLogger(__name__)
        if df_name is not None:
            self.df_name = df_name

        # Load config
        self._config = load_config(self.__version__)

        # Initialize database
        self._initialize_engine(db_engine, db_host, db_name, db_username, db_password)

        # Initialize dataframe
        self.df = self._initialize_df()

        # Fetch table
        if read_on_init:
            self.read()

    def __iter__(self):
        """Return iterator."""
        return self

    def __next__(self):
        """Return the next item."""
        if self._cursor >= len(self.df):
            self._cursor = 0
            raise StopIteration()

        # Grab data
        data = self.df.take([self._cursor]).to_dict(orient='records')[0]

        # Delete internal column
        if 'uuid_in_df' in data.keys():
            del data['uuid_in_df']

        # Post-processes
        data = deserialize_dict_1d(data)

        # Increment
        self._cursor += 1

        return data

    def __len__(self):
        """Return number of records."""
        return len(self.df)

    def _initialize_engine(self,
                           db_engine=None,
                           db_host=None,
                           db_name=None,
                           db_username=None,
                           db_password=None):
        """Initialize DB engine."""
        # Parse
        engine = db_engine if db_engine is not None else self.db_defaults.engine
        username = db_username if db_username is not None else self.db_defaults.username
        password = db_password if db_password is not None else self.db_defaults.password
        host = db_host if db_host is not None else self.db_defaults.host
        database = db_name if db_name is not None else self.db_defaults.database

        # Substitute
        self._config.current_db = {
            'engine': engine,
            'host': host,
            'username': username,
            'password': password,
            'database': database,
        }

        # Connect
        if engine in ['influxdb']:
            if not _extra_supports['influxdb']:
                raise ImportError('Module `influxdb` cannot be not imported')
            hostname = host
            hostport = 8086
            if ':' in host:
                hostname, hostport = host.split(':')
            self._engine = DataFrameClient(hostname, hostport, username, password, database)
        else:
            engine = 'postgresql' if engine == 'timescaledb' \
                else 'mysql' if engine == 'mariadb' \
                else engine
            username_and_password = '' \
                if all([username == '', password == '']) \
                else '{0}:{1}@'.format(username, password)
            self._engine = sqlalchemy.create_engine(
                '{0}://{1}{2}{3}'.format(
                    engine,
                    username_and_password if engine != 'sqlite' else '',
                    '/{0}'.format(host) if engine == 'sqlite' else host,
                    '' if engine == 'sqlite' else '/' + database
                ),
                echo=False
            )
            self._session = scoped_session(sessionmaker(bind=self._engine))

    def _initialize_df(self):
        """Initialize DF."""
        df = pd.concat(
            [pd.Series(name=c['name'],
                       dtype=dtype_string_to_dtype_object(c['dtype'])) for c in self.columns]
            + [pd.Series(name='uuid_in_df', dtype=str)],
            axis=1
        )
        return df

    def _get_column_names(self):
        """Return column names."""
        return [c['name'] for c in self.columns]

    def _get_sql_columns(self):
        return [sql.column(c) for c in self._get_column_names()]

    def _get_uuid_from_item(self, data_in):
        """Return UUID of the given item.

        Args:
            data_in (dict or pandas.Series): dict or Series containing data

        Returns:
            (str): UUID

        """
        item = data_in
        if isinstance(item, pd.Series):
            item = data_in.to_dict()

        pre_hash = ''.join([
            '{:.09f}'.format(item[c['name']])
            if isinstance(item[c['name']], float) else str(item[c['name']])
            for c in self.columns
            if c['name'] in item.keys()
        ])
        pre_hash = pre_hash.encode('utf-8')
        uuid = hashlib.md5(pre_hash).hexdigest()
        return uuid

    def _preprocess_list_of_dicts(self, data_in):
        """Preprocess list of dicts.

        Args:
            data_in (list): list of dicts containing data

        Returns:
            (dict): dict of lists = listed dict

        """
        data = deepcopy(data_in)

        # Serialize (convert list to str)
        self.logger.info('(Preprocess) Serializing...')
        for item in tqdm(data, desc='Serialization', leave=False):
            item = serialize_dict_1d(item)

            # Add df_uuid
            item['uuid_in_df'] = self._get_uuid_from_item(item)

            # Add missing columns
            for column in self.columns:
                column_name, column_dtype = column['name'], column['dtype']
                if column_name not in item.keys():
                    item.update({column_name: None})
                else:
                    dtype_obj = dtype_string_to_dtype_object(column_dtype)
                    if dtype_obj is None:
                        continue
                    if item[column_name] is not None \
                            and not isinstance(item[column_name], dtype_obj):
                        try:
                            item[column_name] = dtype_obj(item[column_name])
                        except ValueError:
                            item[column_name] = np.nan

        # Convert dict to listed dict
        self.logger.info('(Preprocess) Converting...')
        data = dicts_to_listed_dict_2d(data)

        return data

    def _append_listed_dict_to_df(self, data, check_unique=False):
        """Append pre-processed dict to self._df.

        Args:
            data (dict): data to add
            check_unique (bool): if True, it will be checked that the data is unique in the db

        """
        self.df = pd.concat([self.df, pd.DataFrame.from_dict(data)], sort=False)
        if check_unique:
            self.df.drop_duplicates('uuid_in_df', inplace=True)

    def add_data(self, data_in, **kwargs):
        """Add data to db.

        Args:
            data_in (dict): a dict containing data

        """
        self.add_list_of_data([data_in], **kwargs)

    def add_list_of_data(self, data_in, **kwargs):
        """Add list of data to db.

        Args:
            data_in (list): a list of dicts containing data

        """
        data = self._preprocess_list_of_dicts(data_in)
        self.logger.info('Adding data to DB...')
        self._append_listed_dict_to_df(data, **kwargs)
        self.logger.info('Successfully finished adding data to DB')

    def read(self, df_name=None, query=None, where=None, order_by=None, **kwargs):
        """Read data from SQL.

        Args:
            df_name (str): Dataframe name to read
            query (str SQL query or SQLAlchemy Selectable): query to select items
            where (str): query string for filtering items
            order_by (srt): column name to sort by
            **kwargs: kwargs for function `pandas.read_sql_query`
                      or `influxdb.DataFrameClient.query`

        """
        if df_name is not None:
            self.df_name = df_name

        # Create a query
        q = query
        if q is None:
            # Check if the table exits on DB
            if self._config.current_db['engine'] in ['influxdb']:
                if self.df_name not in [entry['name'] for entry in
                                        self._engine.get_list_measurements()]:
                    self.df = self._initialize_df()
                    return
            else:
                if not self._engine.dialect.has_table(self._engine, self.df_name):
                    self.df = self._initialize_df()
                    return

            # Create a sub-query for extracting unique records
            sub_q = sql.select('*', from_obj=sql.table(self.df_name))
            if self._config.current_db['engine'] in ['mysql', 'sqlite']:
                sub_q = sub_q.group_by(sql.column('uuid_in_df'))
            elif self._config.current_db['engine'] in ['postgresql', 'timescaledb']:
                sub_q = sub_q.distinct(sql.column('uuid_in_df'))

            # Create a query
            if self._config.current_db['engine'] in ['influxdb']:
                q = sub_q
            else:
                q = sql.select('*', from_obj=sub_q.alias('temp'))
            if where is not None:
                q = q.where(sql.text(where))
            if order_by is not None:
                q = q.order_by(sql.text(order_by))

        # Read table from DB
        try:
            if self._config.current_db['engine'] in ['influxdb']:
                df = list(self._engine.query(str(q), **kwargs).values())[0]
            else:
                df = pd.read_sql_query(q, self._engine, **kwargs)
        except Exception as e:
            self.logger.warning('Could not execute SQL statement: "{0}" (reason: {1})'.
                                format(str(q), str(e)))
            df = self._initialize_df()

        self.df = df

    def save(self, df=None, remove_duplicates=False, **kwargs):
        """Save data to SQL.

        Args:
            df (pandas.DataFrame): DataFrame to save (if None, self.df will be saved)
            remove_duplicates (bool): if True, duplicated rows will be removed
            **kwargs: kwargs for function `pandas.dataframe.to_sql`
                      or `influxdb.DataFrameClient.write_points`

        """
        dataframe = df if df is not None else self.df
        dataframe.drop_duplicates('uuid_in_df', inplace=True)

        if self._config.current_db['engine'] in ['influxdb']:
            self._engine.write_points(dataframe, self.df_name, **kwargs)
        else:
            if 'index' not in kwargs.keys():
                kwargs.update({'index': False})
            dataframe.to_sql(self.df_name, self._engine, if_exists='append', **kwargs)

        if remove_duplicates:
            if self._config.current_db['engine'] in ['influxdb']:
                logging.warning('Option "remove duplicates" is not supported yet.')
                return

            # Create temporal table
            temp_table_name = self.df_name + '_' + datetime.now().strftime('%s')
            self._initialize_df().to_sql(temp_table_name, self._engine, index=False)

            # Select unique rows and insert into the temporal table
            if self._config.current_db['engine'] in ['mysql', 'sqlite']:
                select = 'select * from "{0}" group by uuid_in_df'.format(self.df_name)
            elif self._config.current_db['engine'] in ['postgresql', 'timescaledb']:
                select = 'select distinct * from "{0}"'.format(self.df_name)
            else:
                raise ValueError('Unsupported engine: {}'.format(self._engine.name))
            q = 'insert into "{0}" {1}'.format(temp_table_name, select)
            self._engine.execute(q)

            # Drop deprecated table if exists
            if self._engine.dialect.has_table(self._engine, self.df_name + '_deprecated'):
                self._engine.execute('drop table "{0}"'.format(self.df_name + '_deprecated'))

            # Deprecate the original table
            self._engine.execute('alter table "{0}" rename to "{1}"'.
                                 format(self.df_name, self.df_name + '_deprecated'))

            # Rename the temporal table
            self._engine.execute('alter table "{0}" rename to "{1}"'.
                                 format(temp_table_name, self.df_name))

            # Drop the deprecated table
            self._engine.execute('drop table "{0}"'.format(self.df_name + '_deprecated'))

    @property
    def df(self):
        """Return df."""
        return self._df

    @df.setter
    def df(self, value):
        """Setter for self.df."""
        if not isinstance(value, pd.DataFrame):
            raise ValueError('Only pandas dataframe is accepted.')

        # Set columns based on the given DF
        if len(self.columns) == 0 and len(value) > 0:
            self.columns = [{'name': c, 'dtype': 'none'} for c in value.columns.to_list()]

        # Add column 'uuid_in_df'
        if 'uuid_in_df' not in value.columns.to_list():
            value['uuid_in_df'] = value.apply(lambda x: self._get_uuid_from_item(x), axis=1)
            self.columns += [{'name': 'uuid_in_df', 'dtype': 'str'}]

        self._df = value

    @property
    def df_name(self):
        """Return df_name."""
        return self._df_name

    @df_name.setter
    def df_name(self, value):
        """Setter for self.df_name."""
        self._df_name = value

    @property
    def columns(self):
        """Return columns of DF."""
        if self._columns is not None:
            return self._columns
        try:
            return self._config[self.df_name]['columns']
        except KeyError:
            return []

    @columns.setter
    def columns(self, value):
        """Set self.columns."""
        if not isinstance(value, list):
            raise ValueError('Columns must be a list of dicts with keys "name" and "dtype".')
        if len(value) < 1:
            raise ValueError('At least one item must be in the list.')
        if not isinstance(value[0], dict) or "name" not in value[0].keys() or "dtype" not in value[0].keys():
            raise ValueError('Columns must be a list of dicts with keys "name" and "dtype".')

        try:
            _ = pd.concat(
                [pd.Series(name=c['name'],
                           dtype=dtype_string_to_dtype_object(c['dtype'])) for c in value]
                + [pd.Series(name='uuid_in_df', dtype=str)],
                axis=1
            )
        except KeyError as e:
            raise ValueError('Unrecognized value: {}'.format(str(e)))

        self._columns = value


class TimeSeriesDBHandler(BaseDBHandler):
    """Time series database handler."""

    __version__ = 'v2'
    db_defaults = load_config(__version__).sql.time_series_defaults
    _df_name = 'time_series_df'
    _columns = None

    def read(self, *args, **kwargs):
        """Read data from SQL.

        Args:
            *args: args
            **kwargs: kwargs

        Returns:
            (pandas.df): table

        """
        # Read table from DB
        if self._config.current_db['engine'] in ['influxdb']:
            super().read(*args, **kwargs)
        else:
            super().read(*args, **kwargs, index_col='timestamp')

        # Post-process
        try:
            df = self.df
            df.reset_index(inplace=True)
            df.rename(columns={'index': 'timestamp'}, inplace=True)
            df['timestamp'] = df['timestamp'].astype(int) / 10**9
        except Exception as e:
            self.logger.warning('An error occurred in post-process: {}'.format(e))
            df = self._initialize_df()

        self.df = df

    def save(self, df=None, **kwargs):
        """Save data to SQL.

        Args:
            df (pandas.DataFrame): dataframe to save. if None, self.df will be saved

        """
        dataframe: pd.DataFrame = df if df is not None else self.df

        # Convert timestamp (float) to datetime
        assert 'timestamp' in dataframe.columns
        df_to_save = dataframe.copy()
        df_to_save['timestamp'] = pd.to_datetime(df_to_save['timestamp'].astype(float), unit='s')
        df_to_save.set_index('timestamp', inplace=True)

        if self._config.current_db['engine'] in ['influxdb']:
            super().save(df=df_to_save)
        else:
            super().save(df=df_to_save, index=True)



"""DB Handlers."""

# Version 2
from .v2 import BaseDBHandler as V2BaseDBHandler
from .v2 import TimeSeriesDBHandler as V2TimeSeriesDBHandler
from .v2 import MetaDBHandler as V2MetaDBHandler
from .v2 import BaseDBSearchEngine as V2BaseSearchEngine
from .v2 import TimeSeriesDBSearchEngine as V2TimeSeriesDBSearchEngine

# Version 3
from .v3 import DBHandler as V3DBHandler
from .v3 import TimeSeriesDBHandler as V3TimeSeriesDBHandler
from .v3 import MetaDBHandler as V3MetaDBHandler
from .v3 import TimeSeriesCassandraDBHandler as V3TimeSeriesCassandraDBHandler
from .v3 import DBSearchEngine as V3DBSearchEngine
from .v3 import TimeSeriesCassandraDBSearchEngine as V3TimeSeriesCassandraDBSearchEngine

# Version 4
from .v4 import DBHandler as V4DBHandler
from .v4 import MetaDBHandler as V4MetaDBHandler
from .v4 import DatabaseIDDBHandler as V4DatabaseIDDBHandler

# Default DB Handler
DBHandler = V4DBHandler



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydtk',
 'pydtk.bin',
 'pydtk.bin.sub_commands',
 'pydtk.builder',
 'pydtk.builder.dbdb',
 'pydtk.db',
 'pydtk.db.v1',
 'pydtk.db.v1.handlers',
 'pydtk.db.v2',
 'pydtk.db.v2.handlers',
 'pydtk.db.v2.search_engines',
 'pydtk.db.v3',
 'pydtk.db.v3.handlers',
 'pydtk.db.v3.search_engines',
 'pydtk.db.v4',
 'pydtk.db.v4.deps',
 'pydtk.db.v4.deps.pql_python37',
 'pydtk.db.v4.deps.pql_python37.pql',
 'pydtk.db.v4.deps.pql_python38',
 'pydtk.db.v4.deps.pql_python38.pql',
 'pydtk.db.v4.engines',
 'pydtk.db.v4.handlers',
 'pydtk.io',
 'pydtk.models',
 'pydtk.models.autoware',
 'pydtk.models.pointcloud',
 'pydtk.models.zstd',
 'pydtk.preprocesses',
 'pydtk.statistics',
 'pydtk.utils']

package_data = \
{'': ['*'], 'pydtk': ['conf/*']}

install_requires = \
['attrdict',
 'bitstring>=3.1.7,<4.0.0',
 'deepmerge>=0.1.1,<0.2.0',
 'fire>=0.3.1,<0.4.0',
 'flatten-dict>=0.3.0,<0.4.0',
 'montydb[lmdb,bson]>=2.3.12,<3.0.0',
 'numpy>=1.16.6,<2.0.0',
 'opencv-python>=4.2.0.34,<5.0.0.0',
 'pandas>=1.0.3,<2.0.0',
 'pymongo>=3.11.3,<4.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pyyaml',
 'six>=1.15.0,<2.0.0',
 'sqlalchemy-migrate',
 'sqlalchemy>=1.3.17,<2.0.0',
 'tinydb>=3.2.1,<4.0.0',
 'tinymongo>=0.2.0,<0.3.0',
 'tqdm>=4.46.1,<5.0.0']

extras_require = \
{'pointcloud': ['pyntcloud'],
 'ros': ['pycryptodomex', 'gnupg', 'rospkg', 'flatdict>=4.0.1,<5.0.0'],
 'zstd': ['pyzstd>=0.15.0,<0.16.0']}

entry_points = \
{'console_scripts': ['analyze_statistics = pydtk.builder.statistic_db:script',
                     'batch_analyze_statistics = '
                     'pydtk.builder.statistic_db:batch_script',
                     'create_meta_db = pydtk.builder.meta_db:script',
                     'make_meta = pydtk.bin.make_meta:main',
                     'pydtk = pydtk.bin.cli:script']}

setup_kwargs = {
    'name': 'pydtk',
    'version': '0.1.11',
    'description': 'A Python toolkit for managing, retrieving and processing data.',
    'long_description': '# Python Dataware Toolkit\n\nA Python toolkit for managing, retrieving, and processing data.\n\n## Installation\nYou can install the toolkit with:\n```bash\n$ pip3 install pydtk\n\n```\n\nIf you want to install the toolkit with extra feature (e.g. support for PointCloud and ROS), \nyou can install it with extra dependencies as follows:\n```bash\n$ pip3 install pydtk[pointcloud,ros]\n\n```\n\nSome PyDTK models require additional packages.  \nPlease refer the following table and install them manually with command `pip install ...`.  \n\n| PyDTK model | Required packages |\n| --- | --- |\n| rosbag.* | ros_numpy (https://github.com/eric-wieser/ros_numpy.git) |\n| pointcloud.PCDModel | pypcd (https://github.com/klintan/pypcd.git) |\n\n\n## Usage\n\nBy using Pydtk, you can load a variety of types of data with a unified interface as shown below.\n\n1. Load DBHandler for retrieving metadata\n```python\nfrom pydtk.db import DBHandler\n\n# Initialize handler (This will read all the metadata from DB on initialization)\nhandler = DBHandler(\n    db_class=\'meta\',\n    db_host=\'./examples/example_db\',\n    base_dir_path=\'./test\'\n)\n\n```\n\n2. Read metadata from db with data selection.\n```python\n# Select by timestamp\nhandler.read(pql=\'start_timestamp > 1420000000 and end_timestamp < 1500000000\')\nprint(handler.data)\n\n# Select by record-id\nhandler.read(pql=\'record_id == regex("test.*")\')\nprint(handler.data)\n\n```\n\n3. Load data from files based on metadata.\n```python\nfrom pydtk.io import BaseFileReader, NoModelMatchedError\n\nreader = BaseFileReader()\n\ntry:\n    for sample in handler:\n        print(\'loading content "{0}" from file "{1}"\'.format(sample[\'contents\'], sample[\'path\']))\n        try:\n            timestamps, data, columns = reader.read(sample)\n            assert print(data)\n        except NoModelMatchedError as e:\n            print(str(e))\n            continue\nexcept EOFError:\n    pass\n```\n\n\n## Documentation\nFor more information about this toolkit, please refer the [document](https://dataware-tools.github.io/pydtk/).\n\n\n## Setup for contribution\nTo improve this toolkit, firstly clone this repository and then \nrun the following command to prepare the environment. \n\n```bash\n$ git clone git@github.com:dataware-tools/pydtk.git --recurse-submodules\n$ poetry install\n\n```\n\nMake sure that [poetry](https://python-poetry.org/) is installed before executing the command.\n\nIf you want to install the toolkit with extra feature (e.g. support for ROS), \nplease specify it with `-E` option.  \nExample (installation with `pointcloud` and `ros` extras):\n```bash\n$ poetry install -E pointcloud -E ros\n\n```\n',
    'author': 'Yusuke Adachi',
    'author_email': 'adachi.yusuke@hdwlab.co.jp',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dataware-tools/python-toolkit.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

# Copyright Toolkit Authors (Yusuke Adachi, Daiki Hayashi)

import contextlib
import os
import re
import sys

import attrdict
from tqdm import tqdm
import yaml


DTYPE_MAP = {
    'string[]': str,
    'string': str,
    'str': str,
    'text': str,
    'int': int,
    'float': float,
    'double': float,
    'boolean': bool,
    'bool': bool,
    'obj': object,
    'object': object,
    'dict': object,
    'datetime': object,
    'none': None
}


def load_config(name):
    """Load config from name.

    Args:
        name (str): name of config (e.g. 'v1')

    Returns:
        (AttrDict): config

    """
    with open(os.path.join(
        os.path.dirname(os.path.dirname(__file__)), 'conf', name + '.yaml'
    ), 'r') as f:
        config = yaml.load(f, Loader=yaml.SafeLoader)
    return attrdict.AttrDict(config)


def tag_filter(tag_list, base_df):
    """Search with tags.

    Args:
        base_df (DataFrame): Search target.
        tag_list (list): List of search tag.

    Returns:
        (DataFrame): Searched DataFrame.

    """
    result_df = base_df
    for tag in tag_list:
        mask = result_df["tag"].apply(lambda x: tag in x)
        result_df = result_df[mask]
    return result_df


def get_record_id_list(file_df):
    """Get record id list from file list.

    Args:
        file_df (DataFrame): Search target.

    Returns:
        record_id_list(list): List of record id

    """
    return file_df["record_id"].unique()


def search_content(content_df, content: str):
    """Search with content.

    Args:
        content_df (DataFrame): Table including content element.
        content (str): Content name as a query.

    Returns:
        (DataFrame): Contents DataFrame including a query content.

    """
    mask = content_df["content"].apply(lambda x: content == x)
    return content_df[mask]


def take_time_and(file_df):
    """Get time.

    Args:
        file_df (DataFrame): Search target.

    Returns:
        start_timestamp (float)
        end_timestamp (float)

    """
    mask = ((file_df["start_timestamp"] != "no info") & (file_df["start_timestamp"] != "Nan"))
    start_timestamp = file_df[mask]["start_timestamp"].max()
    mask = ((file_df["end_timestamp"] != "no info") & (file_df["end_timestamp"] != "Nan"))
    end_timestamp = file_df[mask]["end_timestamp"].min()

    return start_timestamp, end_timestamp


def dict_reg_match(a, b):
    """Match two dicts with regex expression.

    Args:
        a (str, list, dict): dict 1 with regex expression
        b (str, list, dict): dict 2

    Returns:
        (bool): True if the two dict matches

    """
    # type check
    if type(a) is not type(b):
        return False

    if isinstance(a, str):
        return re.fullmatch(a, b) is not None

    if isinstance(a, list):
        return any([any([dict_reg_match(a_, b_) for b_ in b]) for a_ in a])

    if isinstance(a, dict):
        return all([
            any([
                dict_reg_match(av, bv)
                for bk, bv in b.items()
                if re.fullmatch(ak, bk) is not None
            ])
            for ak, av in a.items()
        ])


def serialize_dict_1d(dict_in):
    """Convert list in dict to str.

    Args:
        dict_in (dict): input dict

    Returns:
        (dict): serialized dict

    """
    assert isinstance(dict_in, dict)
    dict_out = dict_in
    for key, value in dict_out.items():
        if isinstance(value, list):
            dict_out.update({key: ';'.join(value)})
    return dict_out


def deserialize_dict_1d(dict_in):
    """Convert str of list to list.

    Args:
        dict_in (dict): input dict

    Returns:
        (dict): deserialized dict

    """
    assert isinstance(dict_in, dict)
    dict_out = dict_in
    for key, value in dict_out.items():
        if isinstance(value, str):
            if ';' in value:
                dict_out.update({key: value.split(';')})
    return dict_out


def dict_to_listed_dict_1d(dict_in):
    """Convert dict to listed dict where each value is a list containing one element.

    Args:
        dict_in (dict): input dict

    Returns:
        (dict): listed dict

    """
    assert isinstance(dict_in, dict)
    dict_out = dict_in
    for key, value in dict_out.items():
        dict_out.update({key: [value]})
    return dict_out


def dicts_to_listed_dict_2d(dicts_in):
    """Convert dicts to listed dict where each value is a list containing multiple elements.

    Args:
        dicts_in (list): list of dicts

    Returns:
        (dict): listed dict

    """
    assert isinstance(dicts_in, list)
    dict_out = dict()
    for item in tqdm(dicts_in, desc='dicts_to_listed_dict_2d', leave=False):
        for key, value in item.items():
            if key not in dict_out.keys():
                dict_out.update({key: [value]})
            else:
                dict_out[key].append(value)
    return dict_out


def listed_dict_to_dict_1d(dict_in):
    """Convert listed dict to dict.

    Args:
        dict_in (dict): input dict (listed dic)

    Returns:
        (dict): dict

    """
    assert isinstance(dict_in, dict)
    dict_out = dict_in
    for key, value in dict_out.items():
        dict_out.update({key: value[0]})
    return dict_out


def dtype_string_to_dtype_object(dtype):
    """Return object corresponds to the input string."""
    try:
        return DTYPE_MAP[dtype]
    except KeyError:
        return object


def _deepmerge_append_list_unique(config, path, base, nxt):
    """A list strategy to append unique elements of nxt."""
    if len(base) == 0:
        return nxt
    if len(nxt) == 0:
        return base
    if isinstance(base[0], dict) or isinstance(nxt[0], dict):
        return [item for item in base] + [item for item in nxt if item not in base]
    if isinstance(base[0], list) or isinstance(nxt[0], list):
        return [item for item in base] + [item for item in nxt if item not in base]
    return list(set(base + nxt))


@contextlib.contextmanager
def smart_open(filename: str = None, mode: str = 'r', *args, **kwargs):
    """Open files and i/o streams transparently.

    Reference:
    https://stackoverflow.com/questions/17602878/how-to-handle-both-with-open-and-sys-stdout-nicely

    Args:
        filename (str): file path (if None, stdout is used)
        mode (str): file open option

    Returns:
        (file-pointer)

    """
    if filename is None or filename == '-':
        if 'r' in mode:
            stream = sys.stdin
        else:
            stream = sys.stdout
        if 'b' in mode:
            fh = stream.buffer
        else:
            fh = stream
        close = False
    else:
        fh = open(filename, mode, *args, **kwargs)
        close = True

    try:
        yield fh
    finally:
        if close:
            try:
                fh.close()
            except AttributeError:
                pass



#!/usr/bin/env python3

import argparse
import json
import logging
import os

from collections import defaultdict
from pydtk.io.reader import BaseFileReader
from pydtk.models import MetaDataModel
from pydtk.utils.utils import load_config, smart_open

config = load_config('v4').bin.make_meta


def make_meta_interactively(template=None):
    """Make metadata with the interactive command."""
    if template is None:
        template = defaultdict(dict)
    meta = defaultdict(dict)
    for key in config.common_item.keys():
        if key in template.keys():
            meta[key] = \
                str(input(f"{config.common_item[key]} [{template[key]}]: ") or template[key])
        else:
            meta[key] = input(f"{config.common_item[key]}: ")
    return meta


def make_meta(file, template=None):
    """Make metadata with a template."""
    meta = template if type(template) is dict else defaultdict(dict)
    meta["path"] = file
    if "contents" in meta.keys():
        meta["contents"] = _get_contents_info(file)
    if "start_timestamp" in meta.keys() and "end_timestamp" in meta.keys():
        meta["start_timestamp"], meta["end_timestamp"] = _get_timestamps_info(file)
    return meta


def _get_contents_info(file_path):
    """Get contents infomation from model.

    Args:
        file_path (str): path to the file

    Returns:
        (dict): contents info

    """
    metadata = MetaDataModel(data={'path': file_path})
    model = BaseFileReader._select_model(metadata)
    contents = model.generate_contents_meta(path=file_path)
    return contents


def _get_timestamps_info(file_path):
    """Get contents infomation from model.

    Args:
        file_path (str): path to the file

    Returns:
        (list): [start_timestamp, end_timestamp]

    """
    metadata = MetaDataModel(data={'path': file_path})
    model = BaseFileReader._select_model(metadata)
    timetamps_info = model.generate_timestamp_meta(path=file_path)
    return timetamps_info


def get_arguments():
    """Parse arguments."""
    parser = argparse.ArgumentParser(description="Metadata maker.")
    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="interactive mode",
    )
    parser.add_argument(
        "--template",
        type=str,
        default=None,
        help="path to json has metadata template",
    )
    parser.add_argument(
        "--file",
        type=str,
        default=None,
        help="file to make metadata",
    )
    parser.add_argument(
        "--out_dir",
        type=str,
        default=None,
        help="output directory",
    )
    return parser.parse_args()


def main():
    """Make metadata."""
    # set logger
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s (%(module)s:%(lineno)d) %(levelname)s: %(message)s",
    )
    args = get_arguments()
    if args.template is not None:
        with open(args.template, "r") as f:
            template = json.load(f)
    else:
        template = None

    if args.interactive:
        meta = make_meta_interactively(template)
        meta_json = input("output json: ")
    else:
        if args.file is None:
            raise ValueError("following arguments are required: --file")
        meta = make_meta(args.file, template)
        if args.out_dir is None:
            meta_json = None
        else:
            meta_json = os.path.join(args.out_dir, os.path.basename(args.file) + ".json")

    with smart_open(meta_json, "wt") as f:
        json.dump(meta, f, indent=4)
    logging.info(f"Dumped: {meta_json}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

from abc import ABC
import cv2
import numpy as np
import os

from pydtk.models import BaseModel, register_model, MetaDataModel
from pydtk.models.csv import CameraTimestampCsvModel


@register_model(priority=1)
class GenericMovieModel(BaseModel, ABC):
    """A generic model for a movie file."""

    _content_type = None    # allow any content-type
    _data_type = None   # allow any data-type
    _file_extensions = ['.mp4']
    _contents = None

    def __init__(self, **kwargs):
        super(GenericMovieModel, self).__init__(**kwargs)

    def _load(self, path, start_timestamp=None, end_timestamp=None, **kwargs):
        """Load a movie file.

        Args:
            path (str): path to a movie file
            start_timestamp (float): timestamp to start loading
            end_timestamp (float): timestamp to end loading

        """
        self.path = path

        # Check existence of the file
        if not os.path.isfile(path):
            raise IOError('No such file: {}'.format(path))

        frames = []
        timestamps = []

        cap = cv2.VideoCapture(path)
        fps = cap.get(cv2.CAP_PROP_FPS)
        width = None
        height = None
        n_channels = None

        while True:
            # Reed the first frame
            ret, frame = cap.read()
            if ret:
                height, width, n_channels = frame.shape
            else:
                break

            # Get timestamp
            frame_current = cap.get(cv2.CAP_PROP_POS_FRAMES)
            sec_current = cap.get(cv2.CAP_PROP_POS_MSEC) / 1000.0

            # Fix sec_current
            if frame_current > 0 and sec_current == 0.0:
                sec_current = frame_current / fps

            # Filter by timestamp
            if start_timestamp is not None and sec_current < start_timestamp:
                continue
            if end_timestamp is not None and sec_current > end_timestamp:
                break

            # Store frames and timestamps
            frames.append(frame)
            timestamps.append(sec_current)

        # Close stream
        cap.release()

        self.data = {
            'timestamps': timestamps,
            'data': frames,
            'fps': fps,
            'width': width,
            'height': height,
            'n_channels': n_channels
        }

    def _save(self, path, **kwargs):
        """Save ndarray data to a mp4 file.

        Args:
            path (str): Path to the output file

        """
        fps = self.data['fps'] if 'fps' in self.data.keys() else 30
        if 'fps' in kwargs.keys():
            fps = kwargs['fps']

        width = self.data['width'] if 'width' in self.data.keys() else 224
        if 'width' in kwargs.keys():
            width = kwargs['width']

        height = self.data['height'] if 'height' in self.data.keys() else 224
        if 'height' in kwargs.keys():
            height = kwargs['height']

        frames = self.data['data']
        fmt = cv2.VideoWriter_fourcc(*'mp4v')
        size = (width, height)
        writer = cv2.VideoWriter(path, fmt, fps, size)

        for frame in frames:
            assert frame.shape[0] == height, \
                'Resolution mismatched! (height {} != {})'.format(height, frame.shape[0])
            assert frame.shape[1] == width, \
                'Resolution mismatched! (width {} != {})'.format(width, frame.shape[1])
            writer.write(frame)

        writer.release()

    def to_ndarray(self):
        """Return data as ndarray."""
        return np.array(self.data['data'])

    @property
    def timestamps(self):
        """Return timestamps as ndarray."""
        return np.array(self.data['timestamps'])

    def load_metadata(self, path):
        """Load and return metadata."""
        metadata = None
        for ext in MetaDataModel._file_extensions:
            metadata_filepath = path + ext
            if os.path.isfile(metadata_filepath):
                metadata = MetaDataModel()
                metadata.load(metadata_filepath)
        if metadata is None:
            raise IOError('No metadata found for file: {}'.format(path))
        return metadata

    def downsample_frames(self, timestamps, target_frame_rate=2.0):
        """Downsample data into the target sampling rate.

        Args:
            timestamps (ndarray): timestamps [sec]

        Returns:
            downsampled_timestamps (ndarray): timestamps [sec]

        """
        downsampled_timestamps = timestamps
        fps_previous_index = 0
        for i, timestamp in enumerate(timestamps):
            fps_current_index = timestamp // (1.0 / float(target_frame_rate))
            if fps_current_index == fps_previous_index:
                downsampled_timestamps[i] = -1
                continue
            else:
                fps_previous_index = fps_current_index

        return downsampled_timestamps

    @classmethod
    def generate_contents_meta(cls, path, content_key='content'):
        """Generate contents metadata.

        Args:
            path (str): File path
            content_key (str): Key of content

        Returns:
            (dict): contents metadata

        """
        # Load file
        cap = cv2.VideoCapture(path)
        fps = cap.get(cv2.CAP_PROP_FPS)
        n_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        # Load the first frame and get information
        ret, frame = cap.read()
        frame_info = {}
        if ret:
            height, width, n_channels = frame.shape
            frame_info.update({
                'height': height,
                'width': width,
                'n_channels': n_channels
            })

        # Get duration
        cap.set(cv2.CAP_PROP_POS_FRAMES, n_frames)
        duration = cap.get(cv2.CAP_PROP_POS_MSEC) / 1000.0

        # Close the file
        cap.release()

        # Generate metadata
        contents = {
            content_key: {
                'tags': ['video'],
                'fps': fps,
                'n_frames': n_frames,
                'duration': duration,
                **frame_info
            }
        }

        return contents

    @classmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        # Load file
        cap = cv2.VideoCapture(path)
        n_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        cap.set(cv2.CAP_PROP_POS_FRAMES, n_frames)
        duration = cap.get(cv2.CAP_PROP_POS_MSEC) / 1000.0

        # Close the file
        cap.release()

        return 0, duration


@register_model(priority=2)
class GenericMovieWithCameraTimestampCsvModel(GenericMovieModel, ABC):
    """A generic model for a movie file with a camera-timestamp-csv."""

    _content_type = None  # allow any content-type
    _data_type = None   # allow any data-type
    _file_extensions = ['.mp4']
    _contents = {'camera/.*': {'tags': ['.*']}}

    def __init__(self, **kwargs):
        super(GenericMovieWithCameraTimestampCsvModel, self).__init__(**kwargs)

    @classmethod
    def _is_loadable(cls, path='', **kwargs):
        """Check data by file format."""
        timestamps_path = path.replace('.mp4', '_timestamps.csv')
        if not os.path.isfile(timestamps_path):
            return False
        return True

    def _load(self, path, start_timestamp=None, end_timestamp=None,
              target_frame_rate=2.0, resize_rate=0.5, raw=False, **kwargs):
        """Load a movie file.

        Args:
            path (str): path to a movie file
            start_timestamp (float): timestamp to start loading
            end_timestamp (float): timestamp to end loading
            target_frame_rate (float): frame rate for downsampling
            resize_rate (float): scale of image resize
            raw (bool): raw image or resized image

        """
        self.path = path

        timestamps_path = path.replace('.mp4', '_timestamps.csv')
        timestamps_metadata = self.load_metadata(timestamps_path)
        timestamps_reader = CameraTimestampCsvModel(metadata=timestamps_metadata)
        timestamps_reader._load(timestamps_path)
        timestamps = timestamps_reader.timestamps

        cap = cv2.VideoCapture(path)
        frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        if not raw:
            timestamps = self.downsample_frames(timestamps, target_frame_rate=target_frame_rate)
        assert frames == len(timestamps)

        frame_ids = np.arange(0, frames)
        frame_ids = frame_ids[np.logical_and(timestamps >= start_timestamp, timestamps < end_timestamp)]
        timestamps = timestamps[frame_ids]

        data = []
        for frame_id in frame_ids:
            cap.set(cv2.CAP_PROP_POS_FRAMES, frame_id)
            ret, frame = cap.read()
            if ret:
                if not raw:
                    frame = cv2.resize(frame, dsize=None, fx=resize_rate,
                                       fy=resize_rate, interpolation=cv2.INTER_LINEAR)
                frame = frame[:, :, ::-1]  # Convert BGR to RGB
                frame = frame.transpose((2, 0, 1))  # Reshape: [H, W, C] -> [C, H, W]
                data.append(frame)

        self.data = {'timestamps': timestamps, 'data': data}

    def _save(self, dir_path, contents=None, **kwargs):
        """Save ndarray data to a csv file.

        Args:
            dir_path (str): path to the output directory

        """
        os.makedirs(dir_path, exist_ok=True)
        base_filename = os.path.basename(self.path)
        timestamps, data = self.timestamps, self.to_ndarray()
        for timestamp, frame in zip(timestamps, data):
            # filename = base_filename + "_" + str(timestamp) + ".jpg"
            filename = "{}_{:.3f}.{}".format(base_filename, timestamp, "jpg")
            output_image = os.path.join(dir_path, filename)
            cv2.imwrite(output_image, frame)

    @classmethod
    def generate_contents_meta(cls, path, content_key='content'):
        """Generate contents metadata.

        Args:
            path (str): File path
            content_key (str): Key of content

        Returns:
            (list): contents metadata

        """
        raise NotImplementedError

    @classmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        raise NotImplementedError
